#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Subpackage to load images and tables."""

import csv
from collections.abc import Sequence
from contextlib import suppress
from io import BytesIO, StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Union

import fsspec
import numpy as np
from numpy.typing import DTypeLike
from PIL import Image

from pyxel.options import global_options

if TYPE_CHECKING:
    import pandas as pd
    import xarray as xr


def load_image(filename: Union[str, Path]) -> np.ndarray:
    """Load a 2D image.

    Parameters
    ----------
    filename : str or Path
        Filename to read an image.
        {.npy, .fits, .txt, .data, .jpg, .jpeg, .bmp, .png, .tiff} are accepted.

    Returns
    -------
    ndarray
        A 2D array.

    Raises
    ------
    FileNotFoundError
        If an image is not found.
    ValueError
        When the extension of the filename is unknown or separator is not found.

    Examples
    --------
    >>> from pyxel import load_image
    >>> load_image("frame.fits")
    array([[-0.66328494, -0.63205819, ...]])

    >>> load_image("https://hostname/folder/frame.fits")
    array([[-0.66328494, -0.63205819, ...]])

    >>> load_image("another_frame.npy")
    array([[-1.10136521, -0.93890239, ...]])

    >>> load_image("rgb_frame.jpg")
    array([[234, 211, ...]])
    """
    # Extract suffix (e.g. '.txt', '.fits'...)
    suffix: str = Path(filename).suffix.lower()

    if isinstance(filename, Path):
        full_filename: Path = filename.expanduser().resolve()
        if not full_filename.exists():
            raise FileNotFoundError(f"Input file '{full_filename}' can not be found.")

        url_path: str = str(full_filename)

    else:
        url_path = filename

    # Define extra parameters to use with 'fsspec'
    extras = {}
    if global_options.cache_enabled:
        url_path = f"simplecache::{url_path}"

        if global_options.cache_folder:
            extras["simplecache"] = {"cache_storage": global_options.cache_folder}

    if suffix.startswith(".fits"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            from astropy.io import fits  # Late import to speed-up general import time

            with BytesIO(file_handler.read()) as content:
                data_2d: np.ndarray = fits.getdata(content)

    elif suffix.startswith(".npy"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            data_2d = np.load(file_handler)

    elif suffix.startswith((".txt", ".data")):
        for sep in ("\t", " ", ",", "|", ";"):
            with suppress(ValueError):
                with fsspec.open(url_path, mode="r", **extras) as file_handler:
                    data_2d = np.loadtxt(file_handler, delimiter=sep, ndmin=2)
                break
        else:
            raise ValueError(f"Cannot find the separator for filename '{url_path}'.")

    elif suffix.startswith((".jpg", ".jpeg", ".png", ".bmp", ".tiff", ".tif")):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            image_2d = Image.open(file_handler)
            image_2d_converted = image_2d.convert("LA")  # RGB to grayscale conversion

        data_2d = np.array(image_2d_converted)[:, :, 0]

    else:
        raise ValueError(
            "Image format not supported. List of supported image formats: "
            ".npy, .fits, .txt, .data, .jpg, .jpeg, .bmp, .png, .tiff, .tif."
        )

    return data_2d


def load_table(
    filename: Union[str, Path],
    header: bool = False,
    dtype: DTypeLike = "float",
) -> "pd.DataFrame":
    """Load a table from a file and returns a pandas dataframe. No header is expected in xlsx.

    Parameters
    ----------
    filename : str or Path
        Filename to read the table.
        {.npy, .xlsx, .csv, .txt., .data} are accepted.
    header : bool, default: False
        Remove the header.
    dtype

    Returns
    -------
    DataFrame

    Raises
    ------
    FileNotFoundError
        If an image is not found.
    ValueError
        When the extension of the filename is unknown or separator is not found.
    """
    # Late import to speedup start-up time
    import pandas as pd

    suffix: str = Path(filename).suffix.lower()

    if isinstance(filename, Path):
        full_filename = Path(filename).expanduser().resolve()
        if not full_filename.exists():
            raise FileNotFoundError(f"Input file '{full_filename}' can not be found.")

        url_path: str = str(full_filename)
    else:
        url_path = filename

    # Define extra parameters to use with 'fsspec'
    extras = {}
    if global_options.cache_enabled:
        url_path = f"simplecache::{url_path}"

        if global_options.cache_folder:
            extras["simplecache"] = {"cache_storage": global_options.cache_folder}

    if suffix.startswith(".npy"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            table = pd.DataFrame(np.load(file_handler), dtype=dtype)

    elif suffix.startswith(".xlsx"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            table = pd.read_excel(
                file_handler,
                header=0 if header else None,
                dtype=float,
            )

    elif suffix.startswith((".txt", ".data", ".csv")):
        # Read file
        with fsspec.open(url_path, mode="r", **extras) as file_handler:
            data: str = file_handler.read()

        valid_delimiters: Sequence[str] = ("\t", " ", ",", "|", ";")
        valid_delimiters_str = "".join(valid_delimiters)

        # Find a delimiter
        try:
            dialect = csv.Sniffer().sniff(data, delimiters=valid_delimiters_str)
        except csv.Error as exc:
            raise ValueError("Cannot find the separator") from exc

        delimiter: str = dialect.delimiter

        if delimiter not in valid_delimiters:
            raise ValueError(f"Cannot find the separator. {delimiter=!r}")

        with StringIO(data) as file_handler:
            if suffix.startswith("csv"):
                table = pd.read_csv(
                    file_handler,
                    delimiter=delimiter,
                    header=0 if header else None,
                    dtype=dtype,
                )
            else:
                table = pd.read_table(
                    file_handler,
                    delimiter=delimiter,
                    header=0 if header else None,
                    dtype=dtype,
                )

    else:
        raise ValueError("Only .npy, .xlsx, .csv, .txt and .data implemented.")

    return table


def load_dataarray(filename: Union[str, Path]) -> "xr.DataArray":
    """Load a ``DataArray`` image.

    Parameters
    ----------
    filename : str of Path

    Returns
    -------
    DataArray
        A multi-dimensional array.

    Raises
    ------
    FileNotFoundError
        If an image is not found.
    """
    if isinstance(filename, Path):
        full_filename: Path = filename.expanduser().resolve()
        if not full_filename.exists():
            raise FileNotFoundError(f"Input file '{full_filename}' can not be found.")

        url_path: str = str(full_filename)

    else:
        url_path = filename

    # Define extra parameters to use with 'fsspec'
    extras = {}
    if global_options.cache_enabled:
        url_path = f"simplecache::{url_path}"

        if global_options.cache_folder:
            extras["simplecache"] = {"cache_storage": global_options.cache_folder}

    import xarray as xr

    with fsspec.open(url_path, mode="rb", **extras) as file_handler:
        data_array = xr.load_dataarray(file_handler)

    return data_array


def load_datacube(filename: Union[str, Path]) -> np.ndarray:
    """Load a 3D datacube.

    Parameters
    ----------
    filename : str or Path
        Filename to read a datacube.
        {.npy} are accepted.

    Returns
    -------
    array : ndarray
        A 3D array.

    Raises
    ------
    FileNotFoundError
        If an image is not found.
    ValueError
        When the extension of the filename is unknown or separator is not found.
    """
    # Extract suffix (e.g. '.txt', '.fits'...)
    suffix: str = Path(filename).suffix.lower()

    if isinstance(filename, Path):
        full_filename: Path = filename.expanduser().resolve()
        if not full_filename.exists():
            raise FileNotFoundError(f"Input file '{full_filename}' can not be found.")

        url_path: str = str(full_filename)

    else:
        url_path = filename

    # Define extra parameters to use with 'fsspec'
    extras = {}
    if global_options.cache_enabled:
        url_path = f"simplecache::{url_path}"

        if global_options.cache_folder:
            extras["simplecache"] = {"cache_storage": global_options.cache_folder}

    if suffix.startswith(".npy"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            data_3d: np.ndarray = np.load(file_handler)
        if np.ndim(data_3d) != 3:
            raise ValueError("Input datacube is not 3-dimensional!")

    else:
        raise ValueError(
            "Image format not supported. List of supported image formats: .npy"
        )

    return data_3d



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel photon generator models."""

from collections.abc import Sequence
from typing import Literal, Optional

import numpy as np

from pyxel.detectors import Detector


def rectangular(
    shape: tuple[int, int],
    level: float,
    object_size: Optional[Sequence[int]] = None,
    object_center: Optional[Sequence[int]] = None,
) -> np.ndarray:
    """Calculate an image of a rectangular object.

    Parameters
    ----------
    shape: tuple
        Shape of the output array.
    level: float
        Flux of photon per pixel.
    object_size: list or tuple, optional
        List or tuple of length 2, integers defining the diameters of the rectangular object
        in vertical and horizontal directions.
    object_center: list or tuple, optional
        List or tuple of length 2, two integers (row and column number),
        defining the coordinates of the center of the rectangular object.

    Returns
    -------
    photon_array: ndarray
        Output numpy array.
    """
    if not object_size:
        raise ValueError(
            "object_size argument should be defined for illumination model"
        )
    if object_size and not len(object_size) == 2:
        raise ValueError("Object size should be a sequence of length 2!")
    if object_center and not len(object_center) == 2:
        raise ValueError("Object size should be a sequence of length 2!")

    photon_array = np.zeros(shape, dtype=float)
    if object_center is not None:
        if not (
            (0 <= object_center[0] <= shape[0]) and (0 <= object_center[1] <= shape[1])
        ):
            raise ValueError('Argument "object_center" should be inside Photon array.')
    else:
        object_center = [int(shape[0] / 2), int(shape[1] / 2)]
    p = object_center[0] - int(object_size[0] / 2)
    q = object_center[1] - int(object_size[1] / 2)
    p0 = int(np.clip(p, a_min=0, a_max=shape[0]))
    q0 = int(np.clip(q, a_min=0, a_max=shape[1]))
    photon_array[slice(p0, p + object_size[0]), slice(q0, q + object_size[1])] = level

    return photon_array


def elliptic(
    shape: tuple[int, int],
    level: float,
    object_size: Optional[Sequence[int]] = None,
    object_center: Optional[Sequence[int]] = None,
) -> np.ndarray:
    """Calculate an image of an elliptic object.

    Parameters
    ----------
    shape: tuple
        Shape of the output array.
    level: float
        Flux of photon per pixel.
    object_size: list or tuple, optional
        List or tuple of length 2, integers defining the diameters of the elliptic object
        in vertical and horizontal directions.
    object_center: list or tuple, optional
        List or tuple of length 2, two integers (row and column number),
        defining the coordinates of the center of the elliptic object.

    Returns
    -------
    photon_array: ndarray
        Output numpy array.
    """
    if not object_size:
        raise ValueError(
            "object_size argument should be defined for illumination model"
        )
    if object_size and not len(object_size) == 2:
        raise ValueError("Object size should be a sequence of length 2!")
    if object_center and not len(object_center) == 2:
        raise ValueError("Object size should be a sequence of length 2!")

    photon_array = np.zeros(shape, dtype=float)
    if object_center is not None:
        if not (
            (0 <= object_center[0] <= shape[0]) and (0 <= object_center[1] <= shape[1])
        ):
            raise ValueError('Argument "object_center" should be inside Photon array.')
    else:
        object_center = [int(shape[0] / 2), int(shape[1] / 2)]
    y, x = np.ogrid[: shape[0], : shape[1]]
    dist_from_center = np.sqrt(
        ((x - object_center[1]) / float(object_size[1] / 2)) ** 2
        + ((y - object_center[0]) / float(object_size[0] / 2)) ** 2
    )
    photon_array[dist_from_center < 1] = level
    return photon_array


def calculate_illumination(
    shape: tuple[int, int],
    level: float,
    option: Literal["uniform", "rectangular", "elliptic"] = "uniform",
    object_size: Optional[Sequence[int]] = None,
    object_center: Optional[Sequence[int]] = None,
) -> np.ndarray:
    """Calculate the array of photons uniformly over the entire array or over a object.

    Parameters
    ----------
    shape: tuple
        Shape of the output array.
    level: float
        Flux of photon per pixel.
    option: str{'uniform', 'elliptic_hole', 'rectangular_hole'}
        A string indicating the type of illumination:
        - ``uniform``
           Uniformly fill the entire array with photon. (Default)
        - ``elliptic_hole``
           Mask with elliptic object.
        - ``rectangular_hole``
           Mask with rectangular object.
    object_size: list or tuple, optional
        List or tuple of length 2, integers defining the diameters of the elliptic or rectangular object
        in vertical and horizontal directions.
    object_center: list or tuple, optional
        List or tuple of length 2, two integers (row and column number),
        defining the coordinates of the center of the elliptic or rectangular object.

    Returns
    -------
    photon_array: ndarray
        Output numpy array.
    """
    if option == "uniform":
        photon_array = np.ones(shape, dtype=float) * level
    elif option == "rectangular":
        photon_array = rectangular(
            shape=shape,
            object_size=object_size,
            object_center=object_center,
            level=level,
        )
    elif option == "elliptic":
        photon_array = elliptic(
            shape=shape,
            object_size=object_size,
            object_center=object_center,
            level=level,
        )
    else:
        raise NotImplementedError

    return photon_array


def illumination(
    detector: Detector,
    level: float,
    option: Literal["uniform", "rectangular", "elliptic"] = "uniform",
    object_size: Optional[Sequence[int]] = None,
    object_center: Optional[Sequence[int]] = None,
    time_scale: float = 1.0,
) -> None:
    """Generate photon uniformly over the entire array or over an elliptic or rectangular object.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    level : float
        Flux of photon per pixel.
    option : str
        A string indicating the type of illumination:
        - ``uniform``
        Uniformly fill the entire array with photon. (Default)
        - ``elliptic``
        Mask with elliptic object.
        - ``rectangular``
        Mask with rectangular object.
    object_size : list or tuple, optional
        List or tuple of length 2, integers defining the diameters of the elliptic or rectangular object
        in vertical and horizontal directions.
    object_center : list or tuple, optional
        List or tuple of length 2, two integers (row and column number),
        defining the coordinates of the center of the elliptic or rectangular object.
    time_scale : float
        Time scale of the photon flux, default is 1 second. 0.001 would be ms.
    """
    shape = (detector.geometry.row, detector.geometry.col)

    photon_array = calculate_illumination(
        shape=shape,
        level=level,
        option=option,
        object_size=object_size,
        object_center=object_center,
    )

    photon_array = photon_array * (detector.time_step / time_scale)

    try:
        detector.photon.array += photon_array
    except ValueError as ex:
        raise ValueError("Shapes of arrays do not match") from ex



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import copy
from pathlib import Path
from typing import Union

import pytest

from pyxel.detectors import (
    APD,
    CCD,
    CMOS,
    MKID,
    APDCharacteristics,
    APDGeometry,
    CCDGeometry,
    Characteristics,
    CMOSGeometry,
    Detector,
    Environment,
    MKIDGeometry,
)

try:
    # Check if library 'asdf' is installed
    import asdf  # noqa: F401
except ImportError:
    WITH_ASDF = False
else:
    WITH_ASDF = True


@pytest.fixture(
    params=(
        "ccd_basic",
        "ccd_100x120",
        "cmos_basic",
        "cmos_100x120",
        "mkid_basic",
        "mkid_100x120",
        "apd_basic",
        "apd_100x120",
    )
)
def detector(request) -> Union[CCD, CMOS, MKID, APD]:
    """Create a valid detector."""
    if request.param == "ccd_basic":
        return CCD(
            geometry=CCDGeometry(row=4, col=5),
            environment=Environment(),
            characteristics=Characteristics(),
        )
    elif request.param == "ccd_100x120":
        return CCD(
            geometry=CCDGeometry(
                row=100,
                col=120,
                total_thickness=123.1,
                pixel_horz_size=12.4,
                pixel_vert_size=34.5,
            ),
            environment=Environment(temperature=100.1),
            characteristics=Characteristics(
                quantum_efficiency=0.1,
                charge_to_volt_conversion=0.2,
                pre_amplification=3.3,
                full_well_capacity=10,
            ),
        )
    elif request.param == "cmos_basic":
        return CMOS(
            geometry=CMOSGeometry(row=4, col=5),
            environment=Environment(),
            characteristics=Characteristics(),
        )
    elif request.param == "cmos_100x120":
        return CMOS(
            geometry=CMOSGeometry(
                row=100,
                col=120,
                total_thickness=123.1,
                pixel_horz_size=12.4,
                pixel_vert_size=34.5,
            ),
            environment=Environment(temperature=100.1),
            characteristics=Characteristics(
                quantum_efficiency=0.1,
                charge_to_volt_conversion=0.2,
                pre_amplification=3.3,
                full_well_capacity=10,
            ),
        )
    elif request.param == "mkid_basic":
        return MKID(
            geometry=MKIDGeometry(row=4, col=5),
            environment=Environment(),
            characteristics=Characteristics(),
        )
    elif request.param == "mkid_100x120":
        return MKID(
            geometry=MKIDGeometry(
                row=100,
                col=120,
                total_thickness=123.1,
                pixel_horz_size=12.4,
                pixel_vert_size=34.5,
            ),
            environment=Environment(temperature=100.1),
            characteristics=Characteristics(
                quantum_efficiency=0.1,
                charge_to_volt_conversion=0.2,
                pre_amplification=3.3,
                full_well_capacity=10,
            ),
        )
    elif request.param == "apd_basic":
        return APD(
            geometry=APDGeometry(row=4, col=5),
            environment=Environment(),
            characteristics=APDCharacteristics(
                roic_gain=1.0, avalanche_gain=2.0, pixel_reset_voltage=3.0
            ),
        )
    elif request.param == "apd_100x120":
        return APD(
            geometry=APDGeometry(
                row=100,
                col=120,
                total_thickness=123.1,
                pixel_horz_size=12.4,
                pixel_vert_size=34.5,
            ),
            environment=Environment(temperature=100.1),
            characteristics=APDCharacteristics(
                quantum_efficiency=0.1,
                full_well_capacity=10,
                adc_bit_resolution=16,
                adc_voltage_range=(0.0, 5.0),
                avalanche_gain=1.0,
                pixel_reset_voltage=12.0,
                roic_gain=4.1,
            ),
        )
    else:
        raise NotImplementedError


def test_equal(detector: Union[CCD, CMOS, MKID, APD]):
    new_detector = copy.deepcopy(detector)

    assert new_detector.geometry is not detector.geometry
    assert new_detector.geometry == detector.geometry

    assert new_detector.environment is not detector.environment
    assert new_detector.environment == detector.environment

    assert new_detector.characteristics is not detector.characteristics
    assert new_detector.characteristics == detector.characteristics

    assert new_detector is not detector
    assert new_detector == detector


def test_to_from_hdf5(detector: Union[CCD, CMOS, MKID, APD], tmp_path: Path):
    """Test methods `Detector.to_hdf5' and `Detector.from_hdf5`."""
    filename = tmp_path / f"{detector.__class__.__name__}.h5"

    # Save to 'hdf5'
    detector.to_hdf5(filename)

    # Load from 'hdf5
    new_detector = Detector.from_hdf5(filename)

    # Comparisons
    assert new_detector.geometry == detector.geometry
    assert new_detector.environment == detector.environment
    assert new_detector.characteristics == detector.characteristics

    assert new_detector == detector


@pytest.mark.skipif(WITH_ASDF is False, reason="Missing library 'asdf'")
def test_to_from_asdf(detector: Union[CCD, CMOS, MKID, APD], tmp_path: Path):
    """Test methods `Detector.to_asdf' and `Detector.from_asdf`."""
    filename = tmp_path / f"{detector.__class__.__name__}.asdf"

    # Save to 'asdf'
    detector.to_asdf(filename)

    # Load from 'asdf'
    new_detector = Detector.from_asdf(filename)

    # Comparisons
    assert new_detector.geometry == detector.geometry
    assert new_detector.environment == detector.environment
    assert new_detector.characteristics == detector.characteristics

    assert new_detector == detector


@pytest.mark.parametrize(
    "filename",
    [
        "detector.h5",
        "detector.hdf",
        "detector.asdf",
        Path("detector.hdf5"),
    ],
)
def test_to_from_filename(
    detector: Union[CCD, CMOS, MKID, APD], tmp_path: Path, filename: str
):
    """Test methods `Detector.load` and `Detector.save`."""
    full_filename: Path = tmp_path / filename

    if full_filename.suffix == ".asdf" and not WITH_ASDF:
        pytest.skip(reason="Missing library 'asdf'")

    # Save
    detector.save(full_filename)

    # Load
    new_detector = Detector.load(full_filename)

    # Comparisons
    assert new_detector.geometry == detector.geometry
    assert new_detector.environment == detector.environment
    assert new_detector.characteristics == detector.characteristics

    assert new_detector == detector



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel CosmiX model to generate charge by ionization."""

import subprocess
from bisect import bisect
from collections.abc import Sequence
from pathlib import Path
from typing import Literal, Optional, Union

import numpy as np
import pandas as pd

from pyxel.detectors import Detector
from pyxel.models.charge_generation.cosmix.particle import Particle
from pyxel.models.charge_generation.cosmix.util import (
    load_histogram_data,
    read_data,
    sampling_distribution,
)


class Simulation:
    """Main class of the program, Simulation contain all the methods to set and run a simulation.

    Parameters
    ----------
    detector : Detector
        Detector object(from :term:`CCD`/:term:`CMOS` library) containing all
        the simulated detector specs.
    """

    def __init__(self, detector: Detector) -> None:
        self.detector = detector
        self.simulation_mode: Optional[
            Literal["cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"]
        ] = None

        self.flux_dist: Optional[np.ndarray] = None
        self.spectrum_cdf: Optional[np.ndarray] = None

        self.energy_loss_data: Optional[
            Literal["stopping", "stepsize", "geant4"]
        ] = None

        self.elec_number_dist = pd.DataFrame()
        self.elec_number_cdf = np.zeros((1, 2))
        self.step_size_dist = pd.DataFrame()
        self.step_cdf = np.zeros((1, 2))
        self.kin_energy_dist = pd.DataFrame()
        self.kin_energy_cdf = np.zeros((1, 2))

        self.data_library = pd.DataFrame()

        self.stopping_power: Optional[np.ndarray] = None

        self.particle: Optional[Particle] = None

        self.particle_type: Optional[
            Literal["proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"]
        ] = None
        self.initial_energy: Optional[Union[int, float, Literal["random"]]] = None
        self.position_ver: Optional[str] = None
        self.position_hor: Optional[str] = None
        self.position_z: Optional[str] = None
        self.angle_alpha: Optional[str] = None
        self.angle_beta: Optional[str] = None
        self.step_length = (
            1.0  # fix, all the other data/parameters should be adjusted to this
        )
        self.energy_cut = 1.0e-5  # MeV
        self.ionization_energy = 3.6

        self.e_num_lst_per_step: list[int] = []
        self.e_energy_lst: list[float] = []
        self.e_pos0_lst: list[float] = []
        self.e_pos1_lst: list[float] = []
        self.e_pos2_lst: list[float] = []
        self.e_vel0_lst: np.ndarray = np.array([])
        self.e_vel1_lst: np.ndarray = np.array([])
        self.e_vel2_lst: np.ndarray = np.array([])

        self.electron_number_from_eloss: list[int] = []
        self.secondaries_from_eloss: list[int] = []
        self.tertiaries_from_eloss: list[int] = []

        self.track_length_lst_per_event: list[float] = []
        self.e_num_lst_per_event: list[int] = []
        self.sec_lst_per_event: list[int] = []
        self.ter_lst_per_event: list[int] = []
        self.edep_per_step: list[float] = []
        self.total_edep_per_particle: list[float] = []
        self.p_energy_lst_per_event: list[float] = []
        self.alpha_lst_per_event: list[float] = []
        self.beta_lst_per_event: list[float] = []

    def parameters(
        self,
        sim_mode: Literal["cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"],
        part_type: Literal[
            "proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"
        ],
        init_energy: Union[int, float, Literal["random"]],
        pos_ver: str,
        pos_hor: str,
        pos_z: str,
        alpha: str,
        beta: str,
        ionization_energy: float = 3.6,
    ) -> None:
        self.simulation_mode = sim_mode
        self.particle_type = part_type
        self.initial_energy = init_energy
        self.position_ver = pos_ver
        self.position_hor = pos_hor
        self.position_z = pos_z
        self.angle_alpha = alpha
        self.angle_beta = beta
        self.ionization_energy = ionization_energy

    def find_smaller_neighbor(self, column: str, value: float) -> float:
        sorted_list: Sequence[float] = sorted(self.data_library[column].unique())
        index = bisect(sorted_list, value) - 1
        if index < 0:
            index = 0
        return sorted_list[index]

    def find_larger_neighbor(self, column: str, value: float) -> float:
        sorted_list: Sequence[float] = sorted(self.data_library[column].unique())
        index = bisect(sorted_list, value)
        if index > len(sorted_list) - 1:
            index = len(sorted_list) - 1
        return sorted_list[index]

    def find_closest_neighbor(self, column: str, value: float) -> float:
        sorted_list: Sequence[float] = sorted(self.data_library[column].unique())
        index_smaller = bisect(sorted_list, value) - 1
        index_larger = bisect(sorted_list, value)

        if index_larger >= len(sorted_list):
            return sorted_list[-1]
        elif (sorted_list[index_larger] - value) < (value - sorted_list[index_smaller]):
            return sorted_list[index_larger]

        return sorted_list[index_smaller]

    def select_stepsize_data(
        self, p_type: str, p_energy: float, p_track_length: float
    ) -> Path:
        df = self.data_library

        distance = self.find_larger_neighbor(column="thickness", value=p_track_length)
        energy = self.find_closest_neighbor(column="energy", value=p_energy)

        df_filtered: pd.DataFrame = df[
            (df.type == p_type) & (df.energy == energy) & (df.thickness == distance)
        ]

        series: pd.Series = df_filtered["path"]
        filename: Path = series.values[0]

        return filename

    def set_stepsize_distribution(self, step_size_file: Path) -> None:
        """TBW.

        :param step_size_file:
        :return:
        .. warning:: EXPERIMENTAL - NOT FINISHED YET
        """
        # # step size distribution in um
        self.step_size_dist = load_histogram_data(
            step_size_file,
            hist_type="step_size",
            skip_rows=4,
            read_rows=10000,
        )

        cum_sum = np.cumsum(self.step_size_dist["counts"])
        cum_sum /= np.max(cum_sum)
        self.step_cdf = np.stack((self.step_size_dist["step_size"], cum_sum), axis=1)

        # # tertiary electron numbers created by secondary electrons
        self.elec_number_dist = load_histogram_data(
            step_size_file,
            hist_type="electron",
            skip_rows=10008,
            read_rows=10000,
        )

        cum_sum_2 = np.cumsum(self.elec_number_dist["counts"])
        cum_sum_2 /= np.max(cum_sum_2)
        self.elec_number_cdf = np.stack(
            (self.elec_number_dist["electron"] - 0.5, cum_sum_2), axis=1
        )

        # # secondary electron spectrum in keV
        # self.kin_energy_dist = load_histogram_data(step_size_file, hist_type='energy', skip_rows=10008, read_rows=200)
        #
        # cum_sum = np.cumsum(self.kin_energy_dist['counts'])
        # cum_sum /= np.max(cum_sum)
        # self.kin_energy_cdf = np.stack((self.kin_energy_dist['energy'], cum_sum), axis=1)

    def event_generation(self) -> bool:
        """Generate an event."""
        track_left = False
        electron_number_per_event = 0
        secondary_per_event = 0
        tertiary_per_event = 0
        geo = self.detector.geometry
        ioniz_energy = self.ionization_energy  # eV

        assert self.simulation_mode is not None
        assert self.particle_type is not None
        assert self.initial_energy is not None
        assert self.spectrum_cdf is not None
        assert self.position_ver is not None
        assert self.position_hor is not None
        assert self.position_z is not None

        self.particle = Particle(
            detector=self.detector,
            simulation_mode=self.simulation_mode,
            particle_type=self.particle_type,
            input_energy=self.initial_energy,
            spectrum_cdf=self.spectrum_cdf,
            starting_pos_ver=self.position_ver,
            starting_pos_hor=self.position_hor,
            starting_pos_z=self.position_z
            # self.angle_alpha, self.angle_beta)
        )

        particle = self.particle
        assert particle.track_length is not None

        self.track_length_lst_per_event += [particle.track_length]

        if self.energy_loss_data == "stepsize":
            # data_filename = self.select_stepsize_data(particle.type, particle.energy, particle.track_length)
            data_filename = self.select_stepsize_data(particle.type, 1000.0, 40.0)
            self.set_stepsize_distribution(data_filename)
            # TODO make a stack of stepsize cdfs and do not load them more than once!!!
        # elif self.energy_loss_data == 'geant4':
        #     pass
        elif self.energy_loss_data == "stopping":
            raise NotImplementedError  # TODO: implement this

        while True:
            if particle.energy <= self.energy_cut:
                break

            # particle.energy is in MeV !
            # particle.deposited_energy is in keV !

            if self.energy_loss_data == "stepsize":
                current_step_size = sampling_distribution(self.step_cdf)  # um
                # e_kin_energy = sampling_distribution(self.kin_energy_cdf)     # keV   TODO
            # elif self.energy_loss_data == 'geant4':
            #     pass
            elif self.energy_loss_data == "stopping":
                raise NotImplementedError  # TODO: implement this

            e_kin_energy = 1.0  # TODO
            particle.deposited_energy = e_kin_energy + ioniz_energy * 1e-3  # keV

            # UPDATE POSITION OF IONIZING PARTICLES
            particle.position[0] += particle.dir_ver * current_step_size  # um
            particle.position[1] += particle.dir_hor * current_step_size  # um
            particle.position[2] += particle.dir_z * current_step_size  # um

            # check if p is still inside detector and have enough energy:
            if (
                not (0.0 < particle.position[0] < geo.vert_dimension)
                or not (0.0 < particle.position[1] < geo.horz_dimension)
                or (0.0 <= particle.position[2] <= (-1 * geo.total_thickness))
                or particle.deposited_energy >= particle.energy * 1e3
            ):
                break

            track_left = True

            particle.energy -= particle.deposited_energy * 1e-3  # MeV

            if self.energy_loss_data == "stepsize":
                # the +1 is the original secondary electron
                electron_number = int(sampling_distribution(self.elec_number_cdf)) + 1
                # electron_number = int(e_kin_energy * 1e3 / ioniz_energy) + 1
            # elif self.energy_loss_data == 'geant4':
            #     electron_number = electron_number_vector[g4_j]
            #     g4_j += 1
            elif self.energy_loss_data == "stopping":
                raise NotImplementedError

            secondary_per_event += 1
            tertiary_per_event += electron_number - 1
            electron_number_per_event += electron_number
            self.e_num_lst_per_step.append(electron_number)

            self.e_energy_lst.append(e_kin_energy * 1e3)  # eV
            self.e_pos0_lst.append(particle.position[0])  # um
            self.e_pos1_lst.append(particle.position[1])  # um
            self.e_pos2_lst.append(particle.position[2])  # um

            self.edep_per_step.append(particle.deposited_energy)  # keV
            particle.total_edep += particle.deposited_energy  # keV

            # save particle trajectory
            particle.trajectory = np.vstack((particle.trajectory, particle.position))
        # END of loop

        if track_left:
            self.total_edep_per_particle.append(particle.total_edep)  # keV
            self.e_num_lst_per_event.append(electron_number_per_event)
            self.sec_lst_per_event.append(secondary_per_event)
            self.ter_lst_per_event.append(tertiary_per_event)

        return False

    def event_generation_geant4(self) -> bool:
        """Generate an event running a geant4 app directly."""
        # error = None
        electron_number_per_event = 0
        secondary_per_event = 0
        tertiary_per_event = 0

        secondaries = 0
        tertiaries = 0

        assert self.simulation_mode is not None
        assert self.particle_type is not None
        assert self.initial_energy is not None
        assert self.spectrum_cdf is not None
        assert self.position_ver is not None
        assert self.position_hor is not None
        assert self.position_z is not None

        self.particle = Particle(
            detector=self.detector,
            simulation_mode=self.simulation_mode,
            particle_type=self.particle_type,
            input_energy=self.initial_energy,
            spectrum_cdf=self.spectrum_cdf,
            starting_pos_ver=self.position_ver,
            starting_pos_hor=self.position_hor,
            starting_pos_z=self.position_z
            # self.angle_alpha, self.angle_beta
        )
        particle = self.particle

        assert particle.track_length is not None
        if particle.track_length < 1.0:
            return True

        self.track_length_lst_per_event.append(particle.track_length)
        self.p_energy_lst_per_event.append(particle.energy)
        self.alpha_lst_per_event.append(particle.alpha)
        self.beta_lst_per_event.append(particle.beta)

        error = subprocess.call(
            [
                "./pyxel/models/charge_generation/cosmix/data/geant4/TestEm18",
                "Silicon",
                particle.type,
                str(particle.energy),
                str(particle.track_length),
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )
        if error != 0:
            return True

        # mat = self.detector.material
        # subprocess.call(['./TestEm18', mat.xxx, particle.type,
        # str(particle.energy), str(particle.track_length)'])

        g4_data_energy_path = Path(__file__).parent.joinpath(
            "data", "geant4", "tars_geant4_energy.data"
        )
        g4energydata = read_data(g4_data_energy_path)  # MeV

        # primary_e_balance = g4energydata[0] * 1.E6
        all_e_loss = g4energydata[1] * 1.0e6
        primary_e_loss = g4energydata[2] * 1.0e6
        secondary_e_loss = g4energydata[3] * 1.0e6
        if all_e_loss > 0.0:
            self.electron_number_from_eloss.append(
                np.floor(all_e_loss / 3.6).astype(int)
            )
            self.secondaries_from_eloss.append(
                np.floor(primary_e_loss / 3.6).astype(int)
            )
            self.tertiaries_from_eloss.append(
                np.floor(secondary_e_loss / 3.6).astype(int)
            )

        g4_data_path = Path(__file__).parent.joinpath(
            "data", "geant4", "tars_geant4.data"
        )
        g4data = read_data(g4_data_path)  # mm (!)

        if g4data.shape == (
            3,
        ):  # alternative running mode, only all electron number without proton step size data
            electron_number_vector: Union[list, np.ndarray] = [g4data[0].astype(int)]
            secondaries = g4data[1].astype(int)
            tertiaries = g4data[2].astype(int)
            step_size_vector: Union[list, np.ndarray] = [0]
        elif g4data.shape == (0,):
            step_size_vector = []  # um
            electron_number_vector = []
        elif g4data.shape == (2,):
            step_size_vector = [g4data[0] * 1e3]  # um
            electron_number_vector = [g4data[1].astype(int)]
        else:
            step_size_vector = g4data[:, 0] * 1e3  # um
            electron_number_vector = g4data[:, 1].astype(int)

        if np.any(electron_number_vector):
            # for j in range(len(step_size_vector)):
            for j in range(len(electron_number_vector)):
                # UPDATE POSITION OF IONIZING PARTICLES
                particle.position[0] += particle.dir_ver * step_size_vector[j]  # um
                particle.position[1] += particle.dir_hor * step_size_vector[j]  # um
                particle.position[2] += particle.dir_z * step_size_vector[j]  # um

                electron_number_per_event += electron_number_vector[j]
                secondary_per_event += secondaries
                tertiary_per_event += tertiaries

                self.e_num_lst_per_step.append(electron_number_vector[j])
                self.e_pos0_lst.append(particle.position[0])  # um
                self.e_pos1_lst.append(particle.position[1])  # um
                self.e_pos2_lst.append(particle.position[2])  # um

                e_kin_energy = 1.0
                self.e_energy_lst.append(e_kin_energy)  # eV

                # self.edep_per_step.append(particle.deposited_energy)    # keV
                # particle.total_edep += particle.deposited_energy        # keV

                # save particle trajectory
                particle.trajectory = np.vstack(
                    (particle.trajectory, particle.position)
                )

            # END of loop

            # self.total_edep_per_particle.append(particle.total_edep)  # keV
            self.e_num_lst_per_event.append(electron_number_per_event)
            self.sec_lst_per_event.append(secondary_per_event)
            self.ter_lst_per_event.append(tertiary_per_event)

        # print('p energy: ', particle.energy, '\ttrack length: ', particle.track_length,
        #       '\telectrons/event: ', electron_number_per_event,
        #       '\tsteps: ', len(step_size_vector), '\terror: ', error)

        return False

    # def _ionization_(self, particle):
    #     """TBW.
    #
    #     :param particle:
    #     :return:
    #     """
    #     geo = self.detector.geometry
    #     ioniz_energy = geo.ionization_energy
    #     let_value = None
    #
    #     # particle.energy is in MeV !
    #     # particle.deposited_energy is in keV !
    #     if self.energy_loss_data == 'let':
    #         let_value = sampling_distribution(self.let_cdf)  # keV/um
    #     elif self.energy_loss_data == 'stopping':
    #         stopping_power = get_yvalue_with_interpolation(self.stopping_power, particle.energy)  # MeV*cm2/g
    #         let_value = 0.1 * stopping_power * geo.material_density  # keV/um
    #
    #     particle.deposited_energy = let_value * self.step_length  # keV
    #
    #     if particle.deposited_energy >= particle.energy * 1e3:
    #         particle.deposited_energy = particle.energy * 1e3
    #
    #     e_kin_energy = 0.1  # eV
    #     electron_number = int(particle.deposited_energy * 1e3 / (ioniz_energy + e_kin_energy))  # eV/eV = 1
    #
    #     self.e_num_lst += [electron_number]
    #     self.e_energy_lst += [e_kin_energy]
    #     self.e_pos0_lst += [particle.position[0]]
    #     self.e_pos1_lst += [particle.position[1]]
    #     self.e_pos2_lst += [particle.position[2]]
    #
    #     # keV
    #     particle.deposited_energy = electron_number * (e_kin_energy + ioniz_energy) * 1e-3
    #     particle.energy -= particle.deposited_energy * 1e-3     # MeV
    #
    #     self.edep_per_step.append(particle.deposited_energy)    # keV
    #     particle.total_edep += particle.deposited_energy        # keV



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Wrapper to create simple graphs using the source extractor package."""
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sep
import xarray as xr
from datatree import DataTree
from matplotlib.patches import Ellipse

from pyxel.detectors import Detector


def show_detector(image_2d: np.ndarray, vmin=0, vmax=100) -> None:
    """Take in the detector object and shows the array values as an image to the user.

    Parameters
    ----------
    image_2d
        2D image array .
    """
    im = plt.imshow(
        image_2d,
        interpolation="nearest",
        cmap="gray",
        vmin=vmin,
        vmax=vmax,
        origin="lower",
    )
    plt.colorbar(im)


def get_background(image_2d: np.ndarray):
    """Get the background of an image using the SEP library.

    Parameters
    ----------
    image_2d
        2d image array.
    """

    return sep.Background(image_2d)


def get_background_image(image_2d: np.ndarray):
    """Get the background of an image and converts it to a 2D-array of the same shape of the original input image.

    Parameters
    ----------
    image_2d
        2D image array .
    """
    bkg = get_background(image_2d)
    return bkg.back()


def get_background_data(image_2d: np.ndarray):
    """Get the background rms array [0], the global average rms [1] and the global average background value [2].

    Parameters
    ----------
    image_2d
        2D image array .
    """
    bkg = get_background(image_2d)
    return bkg.rms(), bkg.globalrms, bkg.globalback


def subtract_background(image_2d: np.ndarray):
    """Return a background subtracted numpy array.

    Parameters
    ----------
    image_2d
        2D image array .
    """
    bkg = get_background(image_2d)
    return image_2d - bkg


def extract_roi(
    image_2d: np.ndarray, thresh: int, minarea: int = 5, name: str = "pixel"
):
    """Return a structured numpy array that gives information on the roi found based on the threshold and minea given.

    Parameters
    ----------
    image_2d : np.ndarray
        2D image array
    thresh : int
        signal level above which signifies a region of interest
    minarea : int
        minimum area of elements required that are above the threshold for the extractor to extract information
    """
    return sep.extract(image_2d, thresh=thresh, segmentation_map=True, minarea=minarea)


def plot_roi(data: np.ndarray, roi) -> None:
    """Plot the input data on a graph and overlays ellipses over the roi's found by the extract function.

    Parameters
    ----------
    data : np.ndarray
        2D image array
    roi : np.ndarray / xarray.Dataset
        structured numpy array or xarray dataset of extracted data
    """
    fig, ax = plt.subplots()
    m, s = np.mean(data), np.std(data)
    im = ax.imshow(
        data,
        interpolation="nearest",
        cmap="gray",
        vmin=m - s,
        vmax=m + s,
        origin="lower",
    )

    # plot an ellipse for each object
    for i in range(len(roi["x"])):
        e = Ellipse(
            xy=(roi["x"][i], roi["y"][i]),
            width=6 * roi["a"][i],
            height=6 * roi["b"][i],
            angle=roi["theta"][i] * 180.0 / np.pi,
        )
        e.set_facecolor("none")
        e.set_edgecolor("red")
        ax.add_artist(e)
    plt.colorbar(im)


def extract_roi_to_xarray(
    detector: Detector,
    array_type: str = "pixel",
    thresh: int = 50,
    minarea: int = 5,
) -> None:
    """Extract the roi data converts it to xarray dataset and saves the information to the final result.

    A warning is generated if the processed data_array is empty.

    Parameters
    ----------
    array_type
    detector : Detector
        Pyxel Detector object.
    thresh : int
        Threshold pixel value above which information from the image array is extracted
    minarea : int
        Minimum area of pixels required that are above the threshold for the extractor to extract information

    Raises
    ------
    ValueError
        If parameter 'array_type' is not 'pixel','signal','image',photon' or 'charge'
    """
    if array_type == "pixel":
        data_2d = detector.pixel.array
    elif array_type == "signal":
        data_2d = detector.signal.array
    elif array_type == "image":
        data_2d = detector.image.array
    elif array_type == "photon":
        data_2d = detector.photon.array
    elif array_type == "charge":
        data_2d = detector.charge.array
    else:
        raise ValueError(
            "Incorrect array_type. Must be one of 'pixel','signal','image',photon' or 'charge'."
        )

    data_2d = np.asarray(data_2d, dtype=float)
    if np.all(data_2d == 0):
        warnings.warn(f"{array_type} data array is empty", stacklevel=2)

    objects, segmap = sep.extract(
        data_2d, thresh=thresh, minarea=minarea, segmentation_map=True
    )

    ds_objects: xr.Dataset = pd.DataFrame(objects).to_xarray()

    num_y, num_x = segmap.shape

    ds_segmap = xr.Dataset()
    ds_segmap["segmap"] = xr.DataArray(
        segmap,
        dims=["image_y", "image_x"],
        coords={"image_y": range(num_y), "image_x": range(num_x)},
    )

    ds = ds_objects.assign_coords(image_y=range(num_y), image_x=range(num_x)).merge(
        ds_segmap
    )

    detector.data["/source_extractor"] = DataTree(ds)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


"""TBW."""

from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Protocol, Union

import pandas as pd
from dask.delayed import Delayed, delayed

from pyxel.outputs import Outputs, ValidFormat, ValidName

if TYPE_CHECKING:
    import xarray as xr

    class SaveToFile(Protocol):
        """TBW."""

        def __call__(self, data: Any, name: str, with_auto_suffix: bool = True) -> Path:
            """TBW."""
            ...


class CalibrationOutputs(Outputs):
    """Collection of methods to save the data buckets from a Detector for a Calibration pipeline.

    Parameters
    ----------
    output_folder : str or Path
        Folder where sub-folder(s) that will be created to save data buckets.
    custom_dir_name : str, optional
        Prefix of the sub-folder name that will be created in the 'output_folder' folder.
        The default prefix is `run_`.
    save_data_to_file : Dict
        Dictionary where key is a 'data bucket' name (e.g. 'detector.photon.array') and value
        is the data format (e.g. 'fits').

        Example:
        {'detector.photon.array': 'fits', 'detector.charge.array': 'hdf', 'detector.image.array':'png'}
    """

    def __init__(
        self,
        output_folder: Union[str, Path],
        custom_dir_name: Optional[str] = "",
        save_data_to_file: Optional[
            Sequence[Mapping[ValidName, Sequence[ValidFormat]]]
        ] = None,
        save_calibration_data: Optional[Sequence[Mapping[str, Sequence[str]]]] = None,
    ):
        super().__init__(
            output_folder=output_folder,
            custom_dir_name=custom_dir_name,
            save_data_to_file=save_data_to_file,
        )

        # Parameter(s) specific for 'Calibration'
        self.save_calibration_data: Optional[
            Sequence[Mapping[str, Sequence[str]]]
        ] = save_calibration_data

    def save_processors(self, processors: pd.DataFrame) -> Sequence[Delayed]:
        """TBW."""
        lst: list[Delayed] = []

        if self.save_data_to_file:
            for _, series in processors.iterrows():
                id_island: int = series["island"]
                id_processor: int = series["id_processor"]
                processor: Delayed = series["processor"]

                # TODO: Create folders ?
                prefix = f"island{id_island:02d}_processor{id_processor:02d}"

                output_filenames: Delayed = delayed(self.save_to_file)(
                    processor=processor, prefix=prefix, with_auto_suffix=False
                )
                lst.append(output_filenames)

        return lst

    def save_calibration_outputs(
        self, dataset: "xr.Dataset", logs: pd.DataFrame
    ) -> None:
        """Save the calibration outputs such as dataset and logs.

        Parameters
        ----------
        dataset: Dataset
        logs: DataFrame
        """

        save_methods: dict[str, SaveToFile] = {"nc": self.save_to_netcdf}

        if self.save_calibration_data is not None:
            dct: Mapping[str, Sequence[str]]
            for dct in self.save_calibration_data:
                first_item, *_ = dct.items()
                obj, format_list = first_item

                if obj == "logs":
                    for formal_file in format_list:
                        if formal_file == "csv":
                            filename = self.output_dir.joinpath("logs.csv")
                            logs.to_csv(filename)
                        elif formal_file == "xlsx":
                            filename = self.output_dir.joinpath("logs.xlsx")
                            logs.to_excel(filename)
                        else:
                            raise NotImplementedError(
                                f"Saving to format {formal_file} not implemented"
                            )

                elif obj == "dataset":
                    if format_list is not None:
                        for out_format in format_list:
                            if out_format not in save_methods:
                                raise ValueError(
                                    "Format " + out_format + " not a valid save method!"
                                )

                            func = save_methods[out_format]
                            func(data=dataset, name=obj)

                else:
                    raise NotImplementedError(f"Object {obj} unknown.")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Observation class and functions."""

import logging
import operator
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Optional, Union

import numpy as np
import xarray as xr
from datatree import DataTree
from tqdm.auto import tqdm

from pyxel import __version__
from pyxel.data_structure import Charge, Image, Photon, Pixel, Signal
from pyxel.pipelines import Processor, ResultId, get_result_id, result_keys
from pyxel.util import set_random_seed

if TYPE_CHECKING:
    from pyxel.detectors import Detector
    from pyxel.exposure import Readout
    from pyxel.outputs import CalibrationOutputs, ExposureOutputs, ObservationOutputs


class Exposure:
    """TBW."""

    def __init__(
        self,
        outputs: "ExposureOutputs",
        readout: "Readout",
        result_type: str = "all",
        pipeline_seed: Optional[int] = None,
    ):
        self.outputs = outputs
        self.readout = readout
        self._result_type: ResultId = get_result_id(result_type)
        self._pipeline_seed = pipeline_seed

    def __repr__(self) -> str:
        cls_name: str = self.__class__.__name__
        return f"{cls_name}<outputs={self.outputs!r}>"

    @property
    def result_type(self) -> ResultId:
        """TBW."""
        return self._result_type

    @result_type.setter
    def result_type(self, value: ResultId) -> None:
        """TBW."""
        self._result_type = value

    @property
    def pipeline_seed(self) -> Optional[int]:
        """TBW."""
        return self._pipeline_seed

    @pipeline_seed.setter
    def pipeline_seed(self, value: Optional[int]) -> None:
        """TBW."""
        self._pipeline_seed = value

    def run_exposure(self, processor: Processor) -> "xr.Dataset":
        """Run an observation pipeline.

        Parameters
        ----------
        processor : Processor

        Returns
        -------
        Dataset
        """
        progressbar = self.readout._num_steps != 1
        y = range(processor.detector.geometry.row)
        x = range(processor.detector.geometry.col)
        times = self.readout.times

        # Unpure changing of processor
        _ = run_exposure_pipeline(
            processor=processor,
            readout=self.readout,
            outputs=self.outputs,
            progressbar=progressbar,
            result_type=self.result_type,
            pipeline_seed=self.pipeline_seed,
        )

        ds: xr.Dataset = processor.result_to_dataset(
            x=x,
            y=y,
            times=times,
            result_type=self.result_type,
        )

        ds.attrs.update({"running mode": "Exposure"})

        return ds

    def run_exposure_new(
        self,
        processor: Processor,
        with_intermediate_steps: bool,
    ) -> DataTree:
        """Run an observation pipeline.

        Parameters
        ----------
        processor : Processor
        with_intermediate_steps : bool

        Returns
        -------
        DataTree
        """
        progressbar = self.readout._num_steps != 1

        # Unpure changing of processor
        data_tree: DataTree = run_pipeline(
            processor=processor,
            readout=self.readout,
            outputs=self.outputs,
            progressbar=progressbar,
            result_type=self.result_type,
            pipeline_seed=self.pipeline_seed,
            with_intermediate_steps=with_intermediate_steps,
        )

        data_tree.attrs["running mode"] = "Exposure"

        return data_tree


# TODO: This function will be deprecated
def run_exposure_pipeline(
    processor: Processor,
    readout: "Readout",
    outputs: Union[
        "CalibrationOutputs", "ObservationOutputs", "ExposureOutputs", None
    ] = None,
    progressbar: bool = False,
    result_type: ResultId = ResultId("all"),  # noqa: B008
    pipeline_seed: Optional[int] = None,
) -> Processor:
    """Run standalone exposure pipeline.

    Parameters
    ----------
    pipeline_seed: int
        Random seed for the pipeline.
    result_type: ResultId
    processor: Processor
    readout: Readout
    outputs: DynamicOutputs
        Sampling outputs.
    progressbar: bool
        Sets visibility of progress bar.

    Returns
    -------
    processor: Processor
    """
    # if isinstance(detector, CCD):
    #    dynamic.non_destructive_readout = False

    with set_random_seed(seed=pipeline_seed):
        num_steps = readout._num_steps
        readout._times[-1]
        readout.time_step_it()

        detector = processor.detector

        detector.set_readout(
            times=readout.times,
            start_time=readout.start_time,
            non_destructive=readout.non_destructive,
        )

        # The detector should be reset before exposure
        detector.empty()

        if progressbar:
            pbar = tqdm(total=num_steps, desc="Readout time: ")

        keys = result_keys(result_type)

        unstacked_result: Mapping[str, list] = {key: [] for key in keys}

        i: int
        time: float
        step: float
        for i, (time, step) in enumerate(
            zip(detector.readout_properties.times, detector.readout_properties.steps)
        ):
            detector.time = time
            detector.time_step = step
            detector.pipeline_count = i

            logging.info("time = %.3f s", time)

            if detector.non_destructive_readout:
                empty_all = False
            else:
                empty_all = True

            detector.empty(empty_all)

            processor.run_pipeline()

            if outputs and detector.read_out:
                outputs.save_to_file(processor)

            for key in keys:
                if key == "data":
                    continue

                unstacked_result[key].append(
                    np.array(operator.attrgetter(key)(detector))
                )

            if progressbar:
                pbar.update(1)

        # TODO: Refactor '.result'. See #524
        processor.result = {
            key: np.stack(unstacked_result[key]) for key in keys if key != "data"
        }

        if progressbar:
            pbar.close()

    return processor


def _extract_datatree(detector: "Detector", keys: Sequence[ResultId]) -> DataTree:
    """Extract data from a detector object into a `DataTree`.

    Parameters
    ----------
    detector : Detector
    keys:
        Bucket(s) to extract (e.g. ["photon", "charge", "pixel", "signal", "image", "data"])

    Returns
    -------
    DataTree

    Notes
    -----
    This function is used internally.

    Examples
    --------
    >>> _extract_datatree(
    ...     detector=detector,
    ...     keys=["photon", "charge", "pixel", "signal", "image", "data"],
    ... )
    DataTree('None', parent=None)
        Dimensions:  (time: 1, y: 100, x: 100)
        Coordinates:
          * time     (time) float64 1.0
          * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
          * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        Data variables:
            photon   (time, y, x) float64 1.515e+04 1.592e+04 ... 1.621e+04 1.621e+04
            charge   (time, y, x) float64 1.515e+04 1.592e+04 ... 1.621e+04 1.621e+04
            pixel    (time, y, x) float64 1.515e+04 1.592e+04 ... 1.621e+04 1.621e+04
            signal   (time, y, x) float64 0.04545 0.04776 0.04634 ... 0.04862 0.04862
            image    (time, y, x) uint32 298 314 304 304 304 314 ... 339 339 328 319 319
    """
    # Get current absolute time
    absolute_time = xr.DataArray(
        [detector.absolute_time],
        dims=["time"],
        attrs={"units": "s"},
    )

    dataset: xr.Dataset = xr.Dataset()

    key: ResultId
    for key in keys:
        if key.startswith("data") or key.startswith("scene"):
            continue

        obj: Union[Photon, Pixel, Image, Signal, Charge] = getattr(detector, key)

        if not isinstance(obj, (Photon, Pixel, Image, Signal, Charge)):
            raise TypeError(
                f"Wrong type from attribute 'detector.{key}'. Type: {type(obj)!r}"
            )

        data_array: xr.DataArray = obj.to_xarray().expand_dims(time=absolute_time)
        data_array.name = "value"

        dataset[key] = data_array

    return DataTree(dataset)


def run_pipeline(
    processor: Processor,
    readout: "Readout",
    outputs: Union[
        "CalibrationOutputs", "ObservationOutputs", "ExposureOutputs", None
    ] = None,
    progressbar: bool = False,
    result_type: ResultId = ResultId("all"),  # noqa: B008
    pipeline_seed: Optional[int] = None,
    with_intermediate_steps: bool = False,
) -> DataTree:
    """Run standalone exposure pipeline.

    Parameters
    ----------
    processor : Processor
    readout : Readout
    outputs : DynamicOutputs
        Sampling outputs.
    progressbar : bool
        Sets visibility of progress bar.
    result_type : ResultId
    pipeline_seed : int
        Random seed for the pipeline.
    with_intermediate_steps : bool

    Returns
    -------
    DataTree
    """
    # if isinstance(detector, CCD):
    #    dynamic.non_destructive_readout = False

    with set_random_seed(seed=pipeline_seed):
        detector = processor.detector

        detector.set_readout(
            times=readout.times,
            start_time=readout.start_time,
            non_destructive=readout.non_destructive,
        )

        # The detector should be reset before exposure
        detector.empty()

        if progressbar:
            pbar = tqdm(
                total=detector.readout_properties.num_steps, desc="Readout time: "
            )

        # Get attributes to extract from 'detector'
        # Example: keys = ['photon', 'charge', 'pixel', 'signal', 'image', 'data']
        keys: Sequence[ResultId] = result_keys(result_type)

        data_tree: DataTree = DataTree()

        i: int
        time: float
        step: float
        for i, (time, step) in enumerate(
            zip(detector.readout_properties.times, detector.readout_properties.steps)
        ):
            detector.readout_properties.time = time
            detector.readout_properties.time_step = step
            detector.readout_properties.pipeline_count = i

            logging.info("time = %.3f s", time)

            # Empty detector (if needed)
            is_destructive_readout: bool = not detector.non_destructive_readout
            detector.empty(is_destructive_readout)

            # Run one pipeline
            processor.run_pipeline(with_intermediate_steps=with_intermediate_steps)

            # Save results in file(s) (if needed)
            if outputs and detector.read_out:
                outputs.save_to_file(processor)

            # Extract data from 'detector' into a 'DataTree'
            partial_datatree: DataTree = _extract_datatree(detector=detector, keys=keys)

            # Concatenate all 'partialtree'
            if data_tree.is_empty:
                data_tree = partial_datatree
            else:
                data_tree = data_tree.combine_first(partial_datatree)

                # Fix dtype of container 'image'. See #652
                image_dtype: np.dtype = data_tree["image"].dtype
                exp_dtype: np.dtype = detector.image.dtype

                if image_dtype != exp_dtype:
                    new_image: xr.DataArray = data_tree["image"].astype(dtype=exp_dtype)
                    data_tree["image"] = new_image

            if progressbar:
                pbar.update(1)

        if with_intermediate_steps:
            # Remove temporary data_tree '/intermediate/last'
            datatree_intermediate: DataTree = detector.data["intermediate"]  # type: ignore
            del datatree_intermediate["last"]

        if "scene" in keys:
            data_tree["/scene"] = detector.scene.data

        if "data" in keys:
            data_tree["/data"] = detector.data

        data_tree.attrs["pyxel version"] = __version__

        if progressbar:
            pbar.close()

    return data_tree



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

import numpy as np
import pytest

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Environment,
    ReadoutProperties,
)
from pyxel.models.data_processing import extract_roi_to_xarray


@pytest.mark.parametrize(
    "array_type,exp_warn",
    [
        ("pixel", "pixel data array is empty"),
        ("signal", "signal data array is empty"),
        ("image", "image data array is empty"),
        ("photon", "photon data array is empty"),
        ("charge", "charge data array is empty"),
    ],
)
def test_extract_roi_to_xarray_empty_array(ccd_10x10: CCD, array_type, exp_warn):
    """Tests empty array warning."""
    with pytest.warns(UserWarning, match=exp_warn):
        extract_roi_to_xarray(detector=ccd_10x10, array_type=array_type)


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


def test_extract_roi_to_xarray_pixel(ccd_10x10: CCD):
    ccd_10x10.pixel.array = np.full(fill_value=1, shape=(10, 10), dtype=float)
    extract_roi_to_xarray(detector=ccd_10x10, array_type="pixel")


def test_extract_roi_to_xarray_signal(ccd_10x10: CCD):
    ccd_10x10.signal.array = np.full(fill_value=1, shape=(10, 10), dtype=float)
    extract_roi_to_xarray(detector=ccd_10x10, array_type="signal")


def test_extract_roi_to_xarray_image(ccd_10x10: CCD):
    ccd_10x10.image.array = np.full(fill_value=1, shape=(10, 10), dtype=np.uint64)
    extract_roi_to_xarray(detector=ccd_10x10, array_type="image")


def test_extract_roi_to_xarray_photon(ccd_10x10: CCD):
    ccd_10x10.photon.array = np.full(fill_value=1, shape=(10, 10), dtype=float)
    extract_roi_to_xarray(detector=ccd_10x10, array_type="photon")


def test_extract_roi_to_xarray_charge(ccd_10x10: CCD):
    ccd_10x10.charge.add_charge_array(
        np.full(fill_value=1, shape=(10, 10), dtype=float)
    )
    extract_roi_to_xarray(detector=ccd_10x10, array_type="charge")


#     #assert np.any(ccd_10x10.pixel.array != 0)
#     """Test to ensure warning isn't triggered for filled array"""


def test_extract_roi_to_xarray_incorrect_array_type(ccd_10x10: CCD, array_type="test"):
    ccd_10x10.pixel.array = np.random.rand(10, 10)
    """Test to ensure warning isn't triggered for filled array"""

    with pytest.raises(ValueError, match=r"Incorrect array_type\. Must be one of"):
        extract_roi_to_xarray(ccd_10x10, array_type="test")



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.
import numpy as np
import pytest
import xarray as xr
from datatree import DataTree
from skimage.draw import line_aa

from pyxel.detectors import (
    CCD,
    CCDGeometry,
    Characteristics,
    Environment,
    ReadoutProperties,
)
from pyxel.models.data_processing import remove_cosmic_rays


def create_image_with_cosmics():
    data = np.zeros(shape=(50, 50), dtype=float)

    # rr, cc = disk((10, 30), 3, shape=None)
    # data[rr, cc] = 200.0
    #
    # rr, cc = disk((30, 10), 5, shape=None)
    # data[rr, cc] = 200.0
    #
    # # data[25, :] = 500.0
    # data[24, :] = 500.0

    rr, cc, val = line_aa(1, 1, 8, 8)
    data[rr, cc] = val * 500.0

    rr, cc, val = line_aa(1, 28, 8, 21)
    data[rr, cc] = val * 300.0

    rr, cc, val = line_aa(25, 25, 26, 26)
    data[rr, cc] = val * 300.0

    # rr, cc, val = circle_perimeter_aa(30, 30, 15)
    # data[rr, cc] = val * 200.0
    #
    rr, cc, val = line_aa(20, 10, 20, 30)
    data[rr, cc] = val * 100.0

    return data


@pytest.fixture
def ccd_50x50() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=50,
            col=50,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


def test_data_ouput(ccd_50x50: CCD):
    """Test output for function 'remove_cosmic_rays'."""
    detector = ccd_50x50
    pixels = create_image_with_cosmics()  # 2d array
    detector.pixel.array = pixels
    remove_cosmic_rays(detector=detector)
    data = detector.data

    assert isinstance(data, DataTree)
    assert "lacosmic" in data

    assert isinstance(data["/lacosmic/cosmic_ray_clean"], xr.DataArray)
    assert isinstance(data["/lacosmic/cosmic_ray_mask"], xr.DataArray)

    assert list(data["/lacosmic/cosmic_ray_clean"].coords.keys()) == ["y", "x", "time"]
    assert list(data["/lacosmic/cosmic_ray_mask"].coords.keys()) == ["y", "x", "time"]

    assert list(data["/lacosmic/cosmic_ray_clean"].coords["time"]) == [0.0]
    assert list(data["/lacosmic/cosmic_ray_mask"].coords["time"]) == [0.0]


def test_ray_removal(ccd_50x50: CCD):
    """Test ray tracing removal."""
    detector = ccd_50x50
    pixels = create_image_with_cosmics()  # 2d array
    detector.pixel.array = pixels
    remove_cosmic_rays(detector=detector)
    data = detector.data

    assert np.sum(data["/lacosmic/cosmic_ray_mask"].data) > 0
    assert np.sum(data["/lacosmic/cosmic_ray_clean"].data) == 0.0



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel CosmiX model to generate charge by ionization."""

import logging
import math
from pathlib import Path
from typing import Literal, Optional, Union

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from pyxel.detectors import Detector

# from pyxel.models.charge_generation.cosmix.plotting import PlottingCosmix
from pyxel.models.charge_generation.cosmix.simulation import Simulation
from pyxel.models.charge_generation.cosmix.util import interpolate_data, read_data
from pyxel.util import set_random_seed

# from astropy import units as u
# TODO: write basic test to check inputs, private function, documentation


# @validators.validate
# @config.argument(name='', label='', units='', validate=)
def cosmix(
    detector: Detector,
    simulation_mode: Optional[
        Literal["cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"]
    ] = None,
    running_mode: Optional[
        Literal["stopping", "stepsize", "geant4", "plotting"]
    ] = None,
    particle_type: Optional[Literal["proton", "alpha", "ion"]] = None,
    initial_energy: Optional[Union[int, float, Literal["random"]]] = None,
    particles_per_second: Optional[float] = None,
    incident_angles: Optional[tuple[str, str]] = None,
    starting_position: Optional[tuple[str, str, str]] = None,
    # step_size_file: str = None,
    # stopping_file: str = None,
    spectrum_file: Optional[str] = None,
    seed: Optional[int] = None,
    ionization_energy: float = 3.6,
    progressbar: bool = True,
) -> None:
    """Apply CosmiX model.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    simulation_mode : literal
        Simulation mode: ``cosmic_rays``, ``radioactive_decay``.
    running_mode : literal
        Mode: ``stopping``, ``stepsize``, ``geant4``, ``plotting``.
    particle_type
        Type of particle: ``proton``, ``alpha``, ``ion``.
    initial_energy : int or float or literal
        Kinetic energy of particle, set `random` for random.
    particles_per_second : float
        Number of particles per second.
    incident_angles : tuple of str
        Incident angles: ``(α, β)``.
    starting_position : tuple of str
        Starting position: ``(x, y, z)``.
    spectrum_file : str
        Path to input spectrum
    seed : int, optional
        Random seed.
    ionization_energy : float
        Mean ionization energy of the semiconductor lattice.
    progressbar : bool
        Progressbar.
    """
    if simulation_mode is None:
        raise ValueError("CosmiX: Simulation mode is not defined")
    if running_mode is None:
        raise ValueError("CosmiX: Running mode is not defined")
    if particle_type is None:
        raise ValueError("CosmiX: Particle type is not defined")
    if particles_per_second is None:
        raise ValueError("CosmiX: Particles per second is not defined")
    if spectrum_file is None:
        raise ValueError("CosmiX: Spectrum is not defined")

    if initial_energy is None:
        initial_energy = "random"  # TODO

    if incident_angles is None:
        incident_angle_alpha, incident_angle_beta = ("random", "random")
    else:
        incident_angle_alpha, incident_angle_beta = incident_angles

    if starting_position is None:
        start_pos_ver, start_pos_hor, start_pos_z = ("random", "random", "random")
    else:
        start_pos_ver, start_pos_hor, start_pos_z = starting_position

    particle_number = int(particles_per_second * detector.time_step)

    with set_random_seed(seed):
        cosmix = Cosmix(
            detector=detector,
            simulation_mode=simulation_mode,
            particle_type=particle_type,
            initial_energy=initial_energy,
            particle_number=particle_number,
            incident_angle_alpha=incident_angle_alpha,
            incident_angle_beta=incident_angle_beta,
            start_pos_ver=start_pos_ver,
            start_pos_hor=start_pos_hor,
            start_pos_z=start_pos_z,
            ionization_energy=ionization_energy,
            progressbar=progressbar,
        )

        # cosmix.set_simulation_mode(simulation_mode)
        # cosmix.set_particle_type(particle_type)                # MeV
        # cosmix.set_initial_energy(initial_energy)              # MeV
        # cosmix.set_particle_number(particle_number)            # -
        # cosmix.set_incident_angles(incident_angles)            # rad
        # cosmix.set_starting_position(starting_position)        # um
        cosmix.set_particle_spectrum(Path(spectrum_file))

        if running_mode == "stepsize":
            cosmix.set_stepsize()
        elif running_mode == "geant4":
            cosmix.set_geant4()
        else:
            raise NotImplementedError(f"{running_mode=} not implemented !")

        cosmix.run()

        # if running_mode == "stopping":
        #     # cosmix.run_mod()          ########
        #     raise NotImplementedError
        #     # cosmix.set_stopping_power(stopping_file)
        #     # cosmix.run()
        # elif running_mode == "stepsize":
        #     cosmix.set_stepsize()
        #     cosmix.run()
        # elif running_mode == "geant4":
        #     cosmix.set_geant4()
        #     cosmix.run()
        # elif running_mode == "plotting":
        #     plot_obj = PlottingCosmix(cosmix, save_plots=True, draw_plots=True)
        #
        #     plot_obj.plot_flux_spectrum()
        #     plot_obj.plot_gaia_vs_gras_hist(normalize=True)
        #
        #     plot_obj.show()
        # else:
        #     raise ValueError


class Cosmix:
    """TBW."""

    def __init__(
        self,
        detector: Detector,
        simulation_mode: Literal[
            "cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"
        ],
        particle_type: Literal[
            "proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"
        ],
        initial_energy: Union[int, float, Literal["random"]],
        particle_number: int,
        incident_angle_alpha: str,
        incident_angle_beta: str,
        start_pos_ver: str,
        start_pos_hor: str,
        start_pos_z: str,
        ionization_energy: float = 3.6,
        progressbar: bool = True,
    ):
        self.simulation_mode = simulation_mode
        self.part_type = particle_type
        self.init_energy = initial_energy
        self.particle_number = particle_number
        self.angle_alpha = incident_angle_alpha
        self.angle_beta = incident_angle_beta
        self.position_ver = start_pos_ver
        self.position_hor = start_pos_hor
        self.position_z = start_pos_z
        self.ionization_energy = ionization_energy
        self._progressbar = progressbar

        self.sim_obj = Simulation(detector)
        self.charge_obj = detector.charge
        self._log = logging.getLogger(__name__)

    # TODO: Is it still used ?
    def set_simulation_mode(
        self,
        sim_mode: Literal["cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"],
    ) -> None:
        self.simulation_mode = sim_mode

    # TODO: Is it still used ?
    def set_particle_type(
        self,
        particle_type: Literal[
            "proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"
        ],
    ) -> None:
        self.part_type = particle_type

    # TODO: Is it still used ?
    def set_initial_energy(self, energy: Union[int, float, Literal["random"]]) -> None:
        self.init_energy = energy

    # TODO: Is it still used ?
    def set_particle_number(self, number: int) -> None:
        self.particle_number = number

    # TODO: Is it still used ?
    def set_incident_angles(self, angles: tuple[str, str]) -> None:
        alpha, beta = angles
        self.angle_alpha = alpha
        self.angle_beta = beta

    # TODO: Is it still used ?
    def set_starting_position(self, start_position: tuple[str, str, str]) -> None:
        position_vertical, position_horizontal, position_z = start_position
        self.position_ver = position_vertical
        self.position_hor = position_horizontal
        self.position_z = position_z

    def set_particle_spectrum(self, file_name: Path) -> None:
        """Set up the particle specs according to a spectrum.

        Parameters
        ----------
        file_name : Path
            Path of the file containing the spectrum.
        """
        spectrum = read_data(file_name)  # nuc/m2*s*sr*MeV
        geo = self.sim_obj.detector.geometry
        detector_area = geo.vert_dimension * geo.horz_dimension * 1.0e-8  # cm2

        spectrum[:, 1] *= 4 * math.pi * 1.0e-4 * detector_area  # nuc/s*MeV

        spectrum_function = interpolate_data(spectrum)

        lin_energy_range = np.arange(
            np.min(spectrum[:, 0]), np.max(spectrum[:, 0]), 0.01
        )
        self.sim_obj.flux_dist = spectrum_function(lin_energy_range)

        cum_sum = np.cumsum(self.sim_obj.flux_dist)
        cum_sum /= np.max(cum_sum)
        self.sim_obj.spectrum_cdf = np.stack((lin_energy_range, cum_sum), axis=1)

    def set_stopping_power(self, stopping_file: Path) -> None:
        self.sim_obj.energy_loss_data = "stopping"
        self.sim_obj.stopping_power = read_data(stopping_file)

    def set_stepsize(self) -> None:
        self.sim_obj.energy_loss_data = "stepsize"
        self.create_data_library()

    def set_geant4(self) -> None:
        self.sim_obj.energy_loss_data = "geant4"

    def create_data_library(self) -> None:
        self.sim_obj.data_library = pd.DataFrame(
            columns=["type", "energy", "thickness", "path"]
        )

        # mat_list = ['Si']

        type_list = [
            "proton"
        ]  # , 'ion', 'alpha', 'beta', 'electron', 'gamma', 'x-ray']
        energy_list = [100.0]  # MeV
        thick_list = [40.0, 50.0, 60.0, 70.0, 100.0]  # um

        # TODO: Fix this. See issue #152
        path = Path(__file__).parent.joinpath("data", "inputs")
        filename_list = [
            "stepsize_proton_100MeV_40um_Si_10k.ascii",
            "stepsize_proton_100MeV_50um_Si_10k.ascii",
            "stepsize_proton_100MeV_60um_Si_10k.ascii",
            "stepsize_proton_100MeV_70um_Si_10k.ascii",
            "stepsize_proton_100MeV_100um_Si_10k.ascii",
        ]

        i = 0
        for pt in type_list:
            for en in energy_list:
                for th in thick_list:
                    data_dict = {
                        "type": pt,
                        "energy": en,
                        "thickness": th,
                        "path": str(Path(path, filename_list[i])),
                    }
                    new_df = pd.DataFrame(data_dict, index=[0])
                    self.sim_obj.data_library = pd.concat(
                        [self.sim_obj.data_library, new_df], ignore_index=True
                    )
                    i += 1

    def run(self) -> None:
        # print("CosmiX - simulation processing...\n")

        self.sim_obj.parameters(
            sim_mode=self.simulation_mode,
            part_type=self.part_type,
            init_energy=self.init_energy,
            pos_ver=self.position_ver,
            pos_hor=self.position_hor,
            pos_z=self.position_z,
            alpha=self.angle_alpha,
            beta=self.angle_beta,
            ionization_energy=self.ionization_energy,
        )

        # Get output folder and create it (if needed)
        out_path = Path("data").resolve()
        out_path.mkdir(parents=True, exist_ok=True)

        self._log.info("Save data in folder '%s'", out_path)

        for k in tqdm(
            range(self.particle_number),
            desc="Cosmix",
            unit=" particle",
            disable=(not self._progressbar),
        ):
            # for k in range(0, self.particle_number):
            err: Optional[bool] = None
            if self.sim_obj.energy_loss_data == "stepsize":  # TODO
                err = self.sim_obj.event_generation()
            elif self.sim_obj.energy_loss_data == "geant4":
                err = self.sim_obj.event_generation_geant4()
            if k % 10 == 0:
                np.save(
                    f"{out_path}/cosmix-e_num_lst_per_event.npy",
                    self.sim_obj.e_num_lst_per_event,
                )
                np.save(
                    f"{out_path}/cosmix-sec_lst_per_event.npy",
                    self.sim_obj.sec_lst_per_event,
                )
                np.save(
                    f"{out_path}/cosmix-ter_lst_per_event.npy",
                    self.sim_obj.ter_lst_per_event,
                )
                np.save(
                    f"{out_path}/cosmix-track_length_lst_per_event.npy",
                    self.sim_obj.track_length_lst_per_event,
                )
                np.save(
                    f"{out_path}/cosmix-p_energy_lst_per_event.npy",
                    self.sim_obj.p_energy_lst_per_event,
                )
                np.save(
                    f"{out_path}/cosmix-alpha_lst_per_event.npy",
                    self.sim_obj.alpha_lst_per_event,
                )
                np.save(
                    f"{out_path}/cosmix-beta_lst_per_event.npy",
                    self.sim_obj.beta_lst_per_event,
                )

                np.save(
                    f"{out_path}/cosmix-e_num_lst_per_step.npy",
                    self.sim_obj.e_num_lst_per_step,
                )
                np.save(f"{out_path}/cosmix-e_pos0_lst.npy", self.sim_obj.e_pos0_lst)
                np.save(f"{out_path}/cosmix-e_pos1_lst.npy", self.sim_obj.e_pos1_lst)
                np.save(f"{out_path}/cosmix-e_pos2_lst.npy", self.sim_obj.e_pos2_lst)

                np.save(
                    f"{out_path}/cosmix-all_e_from_eloss.npy",
                    self.sim_obj.electron_number_from_eloss,
                )
                np.save(
                    f"{out_path}/cosmix-sec_e_from_eloss.npy",
                    self.sim_obj.secondaries_from_eloss,
                )
                np.save(
                    f"{out_path}/cosmix-ter_e_from_eloss.npy",
                    self.sim_obj.tertiaries_from_eloss,
                )
            if err:
                k -= 1

        size = len(self.sim_obj.e_num_lst_per_step)

        self.sim_obj.e_vel0_lst = np.zeros(size)
        self.sim_obj.e_vel1_lst = np.zeros(size)
        self.sim_obj.e_vel2_lst = np.zeros(size)

        self.charge_obj.add_charge(
            particle_type="e",
            particles_per_cluster=np.asarray(self.sim_obj.e_num_lst_per_step),
            init_energy=np.asarray(self.sim_obj.e_energy_lst),
            init_ver_position=np.asarray(self.sim_obj.e_pos0_lst),
            init_hor_position=np.asarray(self.sim_obj.e_pos1_lst),
            init_z_position=np.asarray(self.sim_obj.e_pos2_lst),
            init_ver_velocity=self.sim_obj.e_vel0_lst,
            init_hor_velocity=self.sim_obj.e_vel1_lst,
            init_z_velocity=self.sim_obj.e_vel2_lst,
        )

    def run_mod(self) -> None:
        # TODO: Use `logging`
        print("CosmiX - adding previous cosmic ray signals to image ...\n")

        # TODO: Use `pathlib.Path`
        out_path = "data/"
        e_num_lst_per_step = np.load(out_path + "cosmix-e_num_lst_per_step.npy")
        e_pos0_lst = np.load(out_path + "cosmix-e_pos0_lst.npy")
        e_pos1_lst = np.load(out_path + "cosmix-e_pos1_lst.npy")
        e_pos2_lst = np.load(out_path + "cosmix-e_pos2_lst.npy")

        size = len(e_num_lst_per_step)
        e_energy_lst = np.zeros(size)
        e_vel0_lst = np.zeros(size)
        e_vel1_lst = np.zeros(size)
        e_vel2_lst = np.zeros(size)

        self.charge_obj.add_charge(
            particle_type="e",
            particles_per_cluster=e_num_lst_per_step,
            init_energy=e_energy_lst,
            init_ver_position=e_pos0_lst,
            init_hor_position=e_pos1_lst,
            init_z_position=e_pos2_lst,
            init_ver_velocity=e_vel0_lst,
            init_hor_velocity=e_vel1_lst,
            init_z_velocity=e_vel2_lst,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Util functions to handle random seeds."""

from contextlib import contextmanager
from typing import Optional

import numpy as np


@contextmanager
def set_random_seed(seed: Optional[int] = None):
    """Set temporary seed the random generator.

    Examples
    --------
    import numpy as np

    with set_random_seed(seed=...):
        value = np.random.random()
    """
    if seed is not None:
        previous_state = np.random.get_state()
        try:
            np.random.seed(seed)
            yield
        finally:
            np.random.set_state(previous_state)
    else:
        # Do nothing
        yield



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import numpy as np
import pytest

from pyxel.detectors import (
    CCD,
    CMOS,
    MKID,
    CCDGeometry,
    Characteristics,
    CMOSGeometry,
    Environment,
    MKIDGeometry,
)
from pyxel.models.readout_electronics import dead_time_filter
from pyxel.models.readout_electronics.dead_time import apply_dead_time_filter


@pytest.fixture
def mkid_5x4() -> MKID:
    """Create a valid CCD detector."""
    detector = MKID(
        geometry=MKIDGeometry(
            row=5,
            col=4,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )

    detector.reset()

    return detector


def test_apply_dead_time_filter():
    """Test function 'apply_dead_time_filter."""
    data_2d = np.array(
        [
            [0, 1, 2, 3, 4],
            [5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14],
            [15, 16, 17, 18, 19],
        ]
    )

    result_2d = apply_dead_time_filter(phase_2d=data_2d, maximum_count=12)

    exp_data_2d = np.array(
        [
            [0, 1, 2, 3, 4],
            [5, 6, 7, 8, 9],
            [10, 11, 12, 12, 12],
            [12, 12, 12, 12, 12],
        ]
    )

    np.testing.assert_equal(result_2d, exp_data_2d)


@pytest.mark.parametrize("dead_time", [1.0, 2])
def test_dead_time_filter(mkid_5x4: MKID, dead_time: float):
    """Test function 'dead_time_filter'."""
    dead_time_filter(detector=mkid_5x4)


def test_dead_time_filter_with_ccd():
    """Test model 'dead_time_filter' with a `CCD` detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=5,
            col=4,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )

    with pytest.raises(TypeError, match=r"Expecting an `MKID` object for 'detector'"):
        dead_time_filter(detector=detector)


def test_dead_time_filter_with_cmos():
    """Test model 'dead_time_filter' with a `CMOS` detector."""
    detector = CMOS(
        geometry=CMOSGeometry(
            row=5,
            col=4,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )

    with pytest.raises(TypeError, match=r"Expecting an `MKID` object for 'detector'"):
        dead_time_filter(detector=detector)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import json
from pathlib import Path

import jsonschema
import pytest
from jsonschema import Draft7Validator
from yaml import safe_load


@pytest.fixture
def schema(request: pytest.FixtureRequest) -> dict:
    filename: Path = request.path.parent / "../../static/pyxel_schema.json"
    full_filename = filename.resolve(strict=True)

    with full_filename.open() as fh:
        content = json.load(fh)

    Draft7Validator.check_schema(content)

    return content


@pytest.mark.parametrize(
    "filename",
    [
        "data/calibration.yaml",
        "data/exposure1.yaml",
        "data/exposure2.yaml",
        "data/observation_custom.yaml",
        "data/observation_custom_parallel.yaml",
        "data/observation_product.yaml",
        "data/observation_sequential.yaml",
    ],
)
def test_validate_configuration_file(
    filename: str, schema: dict, request: pytest.FixtureRequest
):
    """Use a JSON schema to validate a YAML configuration file."""
    folder: Path = request.path.parent

    full_filename = folder / filename
    assert full_filename.exists()

    content: str = full_filename.read_text()

    data = safe_load(content)

    jsonschema.validate(data, schema=schema)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel Array class."""

from typing import TYPE_CHECKING, Optional, Union

import numpy as np

from pyxel.util.memory import get_size

if TYPE_CHECKING:
    import xarray as xr


# TODO: Is it possible to move this to `data_structure/__init__.py' ?
# TODO: Does it make sense to force 'self._array' to be read-only ?
#       It could be done with:
#       ... self._array = np.array(value)
#       ... self._array.setflags(write=False)
class Array:
    """Array class."""

    EXP_TYPE: Union[type, np.dtype] = type(None)
    TYPE_LIST: tuple[np.dtype, ...] = ()
    NAME: str = ""
    UNIT: str = ""

    # TODO: Add units ?
    def __init__(self, value: np.ndarray):
        if value.ndim != 2:
            raise ValueError(
                f"Expecting a 2D array. Got an array with {value.ndim} dimensions."
            )

        self.validate_type(value)

        self._array: np.ndarray = value

        self._numbytes = 0

        # TODO: Implement a method to initialized 'self._array' ???

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__

        return f"{cls_name}<shape={self.shape}, dtype={self.dtype}>"

    def __eq__(self, other) -> bool:
        return type(self) is type(other) and np.array_equal(self.array, other.array)

    def validate_type(self, value: np.ndarray) -> None:
        """Validate a value.

        Parameters
        ----------
        value
        """
        cls_name: str = self.__class__.__name__

        if not isinstance(value, np.ndarray):
            raise TypeError(f"{cls_name} array should be a numpy.ndarray")

        if value.dtype not in self.TYPE_LIST:
            exp_type_name: str = str(self.EXP_TYPE)
            raise TypeError(f"Expected type of {cls_name} array is {exp_type_name}.")

    def validate_shape(self, value: np.ndarray) -> None:
        """TBW."""
        cls_name: str = self.__class__.__name__

        if value.shape != self._array.shape:
            raise ValueError(f"Expected {cls_name} array is {self._array.shape}.")

    def __array__(self, dtype: Optional[np.dtype] = None):
        if not isinstance(self._array, np.ndarray):
            raise TypeError("Array not initialized.")
        return np.asarray(self._array, dtype=dtype)

    @property
    def shape(self) -> tuple[int, int]:
        """Return array shape."""
        num_cols, num_rows = self._array.shape
        return num_cols, num_rows

    @property
    def ndim(self) -> int:
        """Return number of dimensions of the array."""
        return self._array.ndim

    @property
    def dtype(self) -> np.dtype:
        """Return array data type."""
        return self._array.dtype

    @property
    def array(self) -> np.ndarray:
        """Two-dimensional numpy array storing the data.

        Only accepts an array with the right type and shape.
        """
        # if self._array is None:
        #     raise ValueError("'array' is not initialized.")

        return self._array

    @array.setter
    def array(self, value: np.ndarray) -> None:
        """Overwrite the two-dimensional numpy array storing the data.

        Only accepts an array with the right type and shape.
        """
        self.validate_type(value)
        self.validate_shape(value)

        # self.type = value.dtype
        self._array = value

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using `Pympler` library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes

    def to_xarray(self, dtype: Optional[np.typing.DTypeLike] = None) -> "xr.DataArray":
        """Convert into a 2D `DataArray` object with dimensions 'y' and 'x'.

        Parameters
        ----------
        dtype : data-type, optional
            Force a data-type for the array.

        Returns
        -------
        DataArray
            2D DataArray objects.

        Examples
        --------
        >>> detector.photon.to_xarray(dtype=float)
        <xarray.DataArray 'photon' (y: 100, x: 100)>
        array([[15149., 15921., 15446., ..., 15446., 15446., 16634.],
               [15149., 15446., 15446., ..., 15921., 16396., 17821.],
               [14555., 14971., 15446., ..., 16099., 16337., 17168.],
               ...,
               [16394., 16334., 16334., ..., 16562., 16325., 16325.],
               [16334., 15978., 16215., ..., 16444., 16444., 16206.],
               [16097., 15978., 16215., ..., 16681., 16206., 16206.]])
        Coordinates:
          * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
          * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        Attributes:
            units:      Ph
        """
        import xarray as xr

        num_rows, num_cols = self.shape
        return xr.DataArray(
            np.array(self.array, dtype=dtype),
            name=self.NAME.lower(),
            dims=["y", "x"],
            coords={"y": range(num_rows), "x": range(num_cols)},
            attrs={"units": self.UNIT, "long_name": self.NAME},
        )

    def plot(self, robust: bool = True) -> None:
        """Plot the array using Matplotlib.

        Parameters
        ----------
        robust : bool, optional
            If True, the colormap is computed with 2nd and 98th percentile
            instead of the extreme values.

        Examples
        --------
        >>> detector.photon.plot()

        .. image:: _static/photon_plot.png
        """
        import matplotlib.pyplot as plt

        arr: xr.DataArray = self.to_xarray()

        arr.plot(robust=robust)
        plt.title(self.NAME)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Tests for full well models."""

import numpy as np
import pytest

from pyxel.detectors import CCD, CCDGeometry, Characteristics, Environment
from pyxel.models.charge_collection import simple_full_well


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )


@pytest.mark.parametrize(
    "fwc",
    [
        pytest.param(
            10,
        ),
        pytest.param(
            None,
        ),
    ],
)
def test_full_well(
    ccd_10x10: CCD,
    fwc: int,
):
    """Test model 'simple_full_well' with valid inputs."""

    detector = ccd_10x10
    detector.characteristics.full_well_capacity = 10
    detector.pixel.array = np.ones((detector.geometry.row, detector.geometry.col)) * 50

    simple_full_well(detector=ccd_10x10, fwc=fwc)

    assert np.max(detector.pixel.array <= 10)


@pytest.mark.parametrize(
    "fwc, exp_exc, exp_error",
    [
        pytest.param(
            -5,
            ValueError,
            "Full well capacity should be a positive number.",
        ),
    ],
)
def test_full_well_bad_inputs(
    ccd_10x10: CCD,
    fwc: int,
    exp_exc,
    exp_error,
):
    """Test model 'simple_full_well' with bad inputs."""
    with pytest.raises(exp_exc, match=exp_error):
        simple_full_well(detector=ccd_10x10, fwc=fwc)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pytest
import xarray as xr
from datatree import DataTree
from datatree.testing import assert_identical

from pyxel.data_structure import Scene


@pytest.fixture
def source() -> xr.Dataset:
    """Create a valid source."""
    return xr.Dataset(
        {
            "x": xr.DataArray([64.97, 11.94, -55.75, -20.66], dims="ref"),
            "y": xr.DataArray([89.62, -129.3, -48.16, 87.87], dims="ref"),
            "weight": xr.DataArray([14.73, 12.34, 14.63, 14.27], dims="ref"),
            "flux": xr.DataArray(
                [
                    [0.1, 0.2, 0.3, 0.4],
                    [0.5, 0.6, 0.7, 0.8],
                    [0.9, 1.0, 1.1, 1.2],
                    [1.3, 1.4, 1.5, 1.6],
                ],
                dims=["ref", "wavelength"],
            ),
        },
        coords={"ref": [0, 1, 2, 3], "wavelength": [336.0, 338.0, 1018.0, 1020.0]},
    )


def test_empty_scene():
    """Tests with an empty Scene."""
    scene = Scene()

    # Test '.data'
    result = scene.data
    assert isinstance(result, DataTree)
    assert_identical(result, DataTree(name="scene"))

    # Test __eq__
    other_scene = Scene()
    assert scene == other_scene


def test_add_source(source: xr.Dataset):
    """Test method 'Scene.add_source'."""
    scene = Scene()

    # Add one source
    scene.add_source(source.copy(deep=True))

    # Check
    data = scene.data
    exp_data = DataTree(name="scene")
    exp_data["/list/0"] = DataTree(source.copy(deep=True))

    assert_identical(data, exp_data)

    # Check __eq__
    another_scene = Scene()
    assert scene != another_scene

    another_scene.add_source(source.copy(deep=True))
    assert scene == another_scene


def test_add_source_twice(source: xr.Dataset):
    """Test method 'Scene.add_source'."""
    scene = Scene()

    # Add inputs
    scene.add_source(source.copy(deep=True))
    scene.add_source(source.copy(deep=True))

    # Check
    data = scene.data
    exp_data = DataTree(name="scene")
    exp_data["/list/0"] = DataTree(source.copy(deep=True))
    exp_data["/list/1"] = DataTree(source.copy(deep=True))

    assert_identical(data, exp_data)

    # Check __eq__
    another_scene = Scene()
    another_scene.add_source(source.copy(deep=True))
    assert scene != another_scene

    another_scene.add_source(source.copy(deep=True))
    assert scene == another_scene


@pytest.mark.parametrize(
    "source, exp_error, exp_message",
    [
        (xr.DataArray(), TypeError, "Expecting a Dataset object for source"),
        (xr.Dataset(), ValueError, "Wrong format for source"),
        (
            xr.Dataset({"x": [1]}, coords={"ref": [1, 2], "wavelength": [1, 2]}),
            ValueError,
            r"Expecting coordinates \'ref\'",
        ),
        (
            xr.Dataset(
                {
                    "x": xr.DataArray([1], dims="ref"),
                    "y": xr.DataArray([1], dims="ref"),
                },
                coords={"ref": [1], "wavelength": [1]},
            ),
            ValueError,
            "Expecting a Dataset with variables",
        ),
    ],
)
def test_add_source_bad_inputs(source, exp_error, exp_message):
    """Test method 'Source.add_source' with bad inputs."""
    scene = Scene()

    with pytest.raises(exp_error, match=exp_message):
        scene.add_source(source)


def test_empty_to_dict_from_dict():
    """Test methods '.to_dict' and '.from_dict'."""
    scene = Scene()

    # Test method '.to_dict'
    dct = scene.to_dict()
    exp_dct = {"/": {"attrs": {}, "coords": {}, "data_vars": {}, "dims": {}}}
    assert dct == exp_dct

    # Test method '.from_dict'
    other_scene = Scene.from_dict(dct)
    assert scene == other_scene


def test_to_dict_from_dict(source: xr.Dataset):
    """Test methods '.to_dict' and '.from_dict'."""
    scene = Scene()
    scene.add_source(source)

    # Test method '.to_dict'
    dct = scene.to_dict()
    exp_dct = {
        "/": {"attrs": {}, "coords": {}, "data_vars": {}, "dims": {}},
        "/list": {"attrs": {}, "coords": {}, "data_vars": {}, "dims": {}},
        "/list/0": {
            "attrs": {},
            "coords": {
                "ref": {"attrs": {}, "data": [0, 1, 2, 3], "dims": ("ref",)},
                "wavelength": {
                    "attrs": {},
                    "data": [336.0, 338.0, 1018.0, 1020.0],
                    "dims": ("wavelength",),
                },
            },
            "data_vars": {
                "flux": {
                    "attrs": {},
                    "data": [
                        [0.1, 0.2, 0.3, 0.4],
                        [0.5, 0.6, 0.7, 0.8],
                        [0.9, 1.0, 1.1, 1.2],
                        [1.3, 1.4, 1.5, 1.6],
                    ],
                    "dims": ("ref", "wavelength"),
                },
                "weight": {
                    "attrs": {},
                    "data": [14.73, 12.34, 14.63, 14.27],
                    "dims": ("ref",),
                },
                "x": {
                    "attrs": {},
                    "data": [64.97, 11.94, -55.75, -20.66],
                    "dims": ("ref",),
                },
                "y": {
                    "attrs": {},
                    "data": [89.62, -129.3, -48.16, 87.87],
                    "dims": ("ref",),
                },
            },
            "dims": {"ref": 4, "wavelength": 4},
        },
    }
    assert dct == exp_dct

    # Test method '.from_dict'
    other_scene = Scene.from_dict(dct)
    assert scene == other_scene



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Simple models to generate charge due to dark current process."""
import warnings
from typing import Optional

import numpy as np

from pyxel.detectors import CMOS
from pyxel.util import set_random_seed


def lambda_e(lambda_cutoff: float) -> float:
    """Compute lambda_e.

    Parameters
    ----------
    lambda_cutoff : int
        Cut-off wavelength of the detector.

    Returns
    -------
    le: float
    """
    lambda_scale = 0.200847413  # um
    lambda_threshold = 4.635136423  # um
    pwr = 0.544071282
    if lambda_cutoff < lambda_threshold:
        le = lambda_cutoff / (
            1
            - ((lambda_scale / lambda_cutoff) - (lambda_scale / lambda_threshold))
            ** pwr
        )
    else:
        le = lambda_cutoff
    return le


def average_dark_current_rule07(
    pitch: float,
    temperature: float,
    cut_off: float,
) -> float:
    """Compute dark current.

    Parameters
    ----------
    pitch : float
        the x and y dimension of the pixel in micrometer
    temperature : float
        the devices temperature
    cut_off : float
        the detector wavelength cut-off in micrometer

    Returns
    -------
    float
        Dark current values. Unit: e-/pixel/s.
    """
    c = 1.16  # activation energy factor
    q = 1.602e-19  # Charge of one electron, unit= A.s
    k = 1.3806504e-23  # Constant of Boltzmann, unit = m2.kg/s2/K

    j0 = 8367  # obtained through fit of experimental data, see paper

    e_c = 1.24 / lambda_e(
        cut_off
    )  # [eV] the optical gap extracted from the mid response cutoff

    rule07 = j0 * np.exp(-c * q * e_c / (k * temperature))  # A/cm3

    amp_to_eps = 6.242e18  # e-/s
    um2_to_cm2 = 1.0e-8
    factor = amp_to_eps * pitch * pitch * um2_to_cm2  # to convert Amp/cm2 in e/pixel/s
    avg_dark_current_rule07 = rule07 * factor

    return avg_dark_current_rule07


def compute_mct_dark_rule07(
    shape: tuple[int, int],
    pitch: float,
    time_step: float,
    temperature: float,
    cut_off: float,
    spatial_noise_factor: Optional[float] = None,
    temporal_noise: bool = True,
) -> np.ndarray:
    """Compute dark current.

    Parameters
    ----------
    shape : tuple
        Output array shape.
    pitch : float
        the x and y dimension of the pixel in micrometer
    time_step : float
        Time step. Unit: s
    temperature : float
        the devices temperature
    cut_off : float
        the detector wavelength cut-off in micrometer
    spatial_noise_factor : float
        Dark current fixed pattern noise factor.
    temporal_noise: bool
        Shot noise.

    Returns
    -------
    ndarray
        Dark current values. Unit: e-
    """

    avg_dark_current_rule07 = average_dark_current_rule07(
        pitch=pitch,
        temperature=temperature,
        cut_off=cut_off,
    )

    if temporal_noise:
        dark_signal_2d_rule07 = np.ones(shape) * avg_dark_current_rule07 * time_step
        dark_current_shot_noise_2d_rule07 = np.random.poisson(
            dark_signal_2d_rule07
        )  # dark current shot noise
        dark_current_2d_rule07 = dark_current_shot_noise_2d_rule07.astype(float)
    else:
        # The number of charge generated with Poisson distribution using rule07 empiric law for lambda
        dark_signal_2d_rule07 = np.ones(shape) * avg_dark_current_rule07 * time_step
        dark_current_2d_rule07 = dark_signal_2d_rule07

    if spatial_noise_factor is not None:
        dark_current_fpn_sigma_rule07 = (
            time_step * avg_dark_current_rule07 * spatial_noise_factor
        )  # sigma of fpn distribution

        dark_current_2d_rule07 = dark_current_2d_rule07 * (
            1 + np.random.lognormal(sigma=dark_current_fpn_sigma_rule07, size=shape)
        )

    if np.isinf(dark_current_2d_rule07).any():
        warnings.warn(
            "Unphysical high value for dark current from fixed pattern noise distribution"
            " will result in inf values. Enable a FWC model to ensure a physical limit.",
            RuntimeWarning,
            stacklevel=2,
        )

    return dark_current_2d_rule07


def dark_current_rule07(
    detector: CMOS,
    cutoff_wavelength: float = 2.5,  # unit: um
    spatial_noise_factor: Optional[float] = None,
    seed: Optional[int] = None,
    temporal_noise: bool = True,
) -> None:
    """Generate charge from dark current process.

    Based on Rule07 paper by W.E. Tennant Journal of Electronic Materials volume 37, pages1406–1410 (2008).

    Parameters
    ----------
    detector : Detector
    cutoff_wavelength : float
        Cutoff wavelength. Unit: um
    spatial_noise_factor : float
        Dark current fixed pattern noise factor.
    seed : int, optional
    temporal_noise : bool, optional
        Shot noise.
    """
    # TODO: investigate on the knee of rule07 for higher 1/le*T values
    if not isinstance(detector, CMOS):
        raise TypeError("Expecting a CMOS object for detector.")
    if not (1.7 <= cutoff_wavelength <= 15.0):
        raise ValueError("'cutoff' must be between 1.7 and 15.0.")

    geo = detector.geometry

    pitch = geo.pixel_vert_size  # assumes a square pitch
    temperature = detector.environment.temperature
    time_step = detector.time_step

    with set_random_seed(seed):
        dark_current_array_rule07 = compute_mct_dark_rule07(
            pitch=pitch,
            temperature=temperature,
            cut_off=cutoff_wavelength,
            shape=geo.shape,
            time_step=time_step,
            spatial_noise_factor=spatial_noise_factor,
            temporal_noise=temporal_noise,
        )

    detector.charge.add_charge_array(dark_current_array_rule07)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

from collections import abc
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Optional, Union

import h5py as h5
import numpy as np
import pandas as pd
import xarray as xr

from pyxel import __version__

ATTRIBUTES: Mapping[str, Mapping[str, str]] = {
    "photon": {"name": "Photon", "unit": "photon/s"},
    "charge": {"name": "Charge", "unit": "electron"},
    "pixel": {"name": "Pixel", "unit": "electron"},
    "signal": {"name": "Signal", "unit": "volt"},
    "image": {"name": "Image", "unit": "adu"},
}


def _store(
    h5file: h5.File,
    name: str,
    dct: Mapping[
        str, Union[int, float, pd.DataFrame, pd.Series, xr.Dataset, np.ndarray, dict]
    ],
    attributes: Optional[Mapping[str, Mapping[str, str]]] = None,
) -> None:
    """Write data into a new HDF5 group.

    Parameters
    ----------
    h5file : h5.File
        Writeable HDF5 file object.
    name : str
        Name of the HDF5 group to create. (e.g. '/' or '/geometry')
    dct : dict
        Data to write into a HDF5 dataset.
    attributes : dict
        Attributes to store.
    """
    for key, value in dct.items():
        key = key.removeprefix("/")

        new_name = f"{name}/{key}"

        if isinstance(value, (int, float)) or value is None:
            if value is None:
                value = np.nan

            dataset: h5.Dataset = h5file.create_dataset(
                name=new_name, data=value, shape=()
            )

            if attributes is not None and key in attributes:
                dataset.attrs.update(attributes[key])

        elif isinstance(value, pd.DataFrame):
            new_group: h5.Group = h5file.create_group(name=new_name)
            new_group.attrs["type"] = "DataFrame"

            _store(h5file, name=new_name, dct=value.to_dict(orient="series"))

        elif isinstance(value, (pd.Series, np.ndarray, abc.Sequence)):
            dataset = h5file.create_dataset(name=new_name, data=value)

            if attributes is not None and key in attributes:
                dataset.attrs.update(attributes[key])

        elif isinstance(value, xr.Dataset):
            if value:
                _store(h5file, name=new_name, dct=value.to_dict())
            else:
                # Do nothing
                pass

        elif isinstance(value, dict):
            new_group = h5file.create_group(name=new_name)

            if attributes is not None and key in attributes:
                new_group.attrs.update(attributes[key])

            _store(h5file, name=new_name, dct=value)

        else:
            raise NotImplementedError


def to_hdf5(filename: Union[str, Path], dct: Mapping[str, Any]) -> None:
    """Write data to a HDF5 file."""
    if dct["version"] != 1:
        raise NotImplementedError

    with h5.File(filename, mode="w") as h5file:
        # Write main attributes
        h5file.attrs["version"] = dct["version"]
        h5file.attrs["type"] = dct["type"]
        h5file.attrs["pyxel-version"] = __version__

        _store(h5file, name="/geometry", dct=dct["properties"]["geometry"])
        _store(h5file, name="/environment", dct=dct["properties"]["environment"])
        _store(
            h5file, name="/characteristics", dct=dct["properties"]["characteristics"]
        )

        _store(h5file, name="/data", dct=dct["data"], attributes=ATTRIBUTES)


def _load(
    h5file: h5.File, name: str
) -> Union[None, int, float, str, Mapping[str, Any], np.ndarray, pd.DataFrame]:
    """Write data from a HDF5 group.

    Parameters
    ----------
    h5file : h5.File
        Readable HDF5 file object.
    name : str
        Name of the HDF5 group to read. (e.g. '/' or '/geometry')

    Returns
    -------
    dict
        Data to read from a HDF5 dataset.
    """
    dataset: Union[h5.Dataset, h5.Group] = h5file[name]

    if isinstance(dataset, h5.Group):
        dct = {}
        for key, _ in h5file[name].items():
            result = _load(h5file, name=f"{name}/{key}")

            dct[key] = result

        if name.endswith("frame"):
            return pd.DataFrame.from_dict(dct)
        else:
            return dct

    elif isinstance(dataset, h5.Dataset):
        if dataset.ndim == 0:
            value = np.array(dataset, dtype=dataset.dtype)

            if dataset.dtype == np.dtype(object):
                value_converted = value.astype(str)
                return str(value_converted)
            elif np.isnan(value):
                return None
            elif np.issubdtype(dataset.dtype, np.integer):
                return int(value)
            elif np.issubdtype(dataset.dtype, np.floating):
                return float(value)
            else:
                raise TypeError
        else:
            if dataset.dtype == np.dtype(object):
                return np.array(dataset, dtype=str)
            else:
                return np.array(dataset)
    else:
        raise NotImplementedError


@contextmanager
def from_hdf5(filename: Union[str, Path]) -> Iterator[Mapping[str, Any]]:
    """Read data from a HDF5 file."""
    dct = {}
    with h5.File(filename, mode="r") as h5file:
        # Read main attributes
        dct.update(h5file.attrs)

        # TODO: Use a JSON schema to validate 'dct'
        if "version" not in dct or "type" not in dct:
            raise ValueError("Missing 'version' and/or 'type' !")

        version: int = dct["version"]
        if version != 1:
            raise NotImplementedError

        # Get properties
        properties = {}

        name: str
        for name in ("geometry", "environment", "characteristics"):
            properties[name] = _load(h5file, name=f"/{name}")

        dct["properties"] = properties

        # Get data
        data = {}
        for name in h5file["/data"]:
            data[name] = _load(h5file, name=f"/data/{name}")

        dct["data"] = data

        yield dct



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Sampling detector properties class."""

from collections.abc import Sequence
from typing import Union

import numpy as np

from pyxel.exposure.readout import calculate_steps


class ReadoutProperties:
    """Readout sampling detector properties.

    Parameters
    ----------
    times : Sequence[Number]
        A sequence of numeric values representing the sampling times for the readout simulation.
    start_time : float, optional. Default: 0.0
        A float representing the starting time of the readout simulation.
        The readout time(s) should be greater that this ``start_time``.
    non_destructive : bool, optional. Default: False
        A boolean flag indicating whether the readout simulation is non-destructive.
        If set to ``True``, the readout process will not modify the underlying data.

    Examples
    --------
    >>> readout_properties = ReadoutProperties(times=[1, 2, 4, 7, 10], start_time=0.0)
    >>> readout_properties.times
    array([ 1.,  2.,  4.,  7., 10.])
    >>> readout_properties.steps
    array([0., 1., 2., 3., 3.])
    >>> readout_properties.num_steps
    5
    >>> readout_properties.absolute_time
    0.0
    """

    def __init__(
        self,
        times: Union[Sequence[float], np.ndarray],
        start_time: float = 0.0,
        non_destructive: bool = False,
    ):
        times_1d = np.array(times, dtype=float)

        if times_1d.ndim != 1:
            raise ValueError("Readout times must be 1D")
        elif times_1d[0] == 0:
            raise ValueError("Readout times should be non-zero values.")
        elif start_time >= times_1d[0]:
            raise ValueError("Readout times should be greater than start time.")
        elif not np.all(np.diff(times_1d) > 0):
            raise ValueError("Readout times must be strictly increasing")

        steps = calculate_steps(times=times_1d, start_time=start_time)

        self._times: np.ndarray = times_1d
        self._steps: np.ndarray = steps

        self._times.flags.writeable = False
        self._steps.flags.writeable = False

        # Fixed at beginning
        self._num_steps: int = len(self._steps)
        self._start_time: float = start_time
        self._end_time: float = self._times[-1]
        self._non_destructive: bool = non_destructive
        self._times_linear: bool = bool(np.all(self._steps == self._steps[0]))

        # Changing
        self._time: float = 0.0
        self._time_step: float = 1.0
        self._read_out: bool = True
        self._pipeline_count: int = 0

    @property
    def times(self) -> np.ndarray:
        """Return readout times."""
        return self._times

    @property
    def steps(self) -> np.ndarray:
        """Return time steps between consecutive readout times."""
        return self._steps

    @property
    def num_steps(self) -> int:
        """TBW."""
        return self._num_steps

    @property
    def start_time(self) -> float:
        """TBW."""
        return self._start_time

    @start_time.setter
    def start_time(self, value: float) -> None:
        """TBW."""
        self._start_time = value

    @property
    def end_time(self) -> float:
        """TBW."""
        return self._end_time

    @property
    def non_destructive(self) -> bool:
        """TBW."""
        return self._non_destructive

    @property
    def times_linear(self) -> bool:
        """TBW."""
        return self._times_linear

    @property
    def time(self) -> float:
        """TBW."""
        return self._time

    @time.setter
    def time(self, value: float) -> None:
        """TBW."""
        self._time = value

    @property
    def absolute_time(self) -> float:
        """TBW."""
        return self._start_time + self._time

    @property
    def time_step(self) -> float:
        """TBW."""
        return self._time_step

    @time_step.setter
    def time_step(self, value: float) -> None:
        """TBW."""
        self._time_step = value

    @property
    def read_out(self) -> bool:
        """TBW."""
        return self._read_out

    @read_out.setter
    def read_out(self, value: bool) -> None:
        """TBW."""
        self._read_out = value

    @property
    def pipeline_count(self) -> int:
        """TBW."""
        return self._pipeline_count

    @pipeline_count.setter
    def pipeline_count(self, value: int) -> None:
        """TBW."""
        self._pipeline_count = value

    @property
    def is_first_readout(self) -> bool:
        """Check if this is the first readout time."""
        return bool(self.pipeline_count == 0)

    @property
    def is_last_readout(self) -> bool:
        """Check if this is the last readout time."""
        return bool(self.pipeline_count == (self.num_steps - 1))



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import logging
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

import numpy as np
from dask.delayed import Delayed

from pyxel.calibration import (
    Algorithm,
    ArchipelagoDataTree,
    CalibrationMode,
    DaskBFE,
    DaskIsland,
    FitRange3D,
    Island,
    MyArchipelago,
    to_fit_range,
)
from pyxel.calibration.fitting import ModelFitting
from pyxel.calibration.fitting_datatree import ModelFittingDataTree
from pyxel.exposure import Readout
from pyxel.observation import ParameterValues
from pyxel.pipelines import FitnessFunction, Processor, ResultId, get_result_id

if TYPE_CHECKING:
    import pandas as pd
    import xarray as xr
    from datatree import DataTree

    from pyxel.outputs import CalibrationOutputs


def to_path_list(values: Sequence[Union[str, Path]]) -> Sequence[Path]:
    """Convert a sequence of ``Path``-like into a sequence of ``Path``."""
    return [Path(obj).resolve() for obj in values]


class Calibration:
    """TBW."""

    def __init__(
        self,
        outputs: "CalibrationOutputs",
        target_data_path: Sequence[Union[str, Path]],
        fitness_function: FitnessFunction,
        algorithm: Algorithm,
        parameters: Sequence[ParameterValues],
        readout: Optional["Readout"] = None,
        mode: Literal["pipeline", "single_model"] = "pipeline",
        result_type: Literal["image", "signal", "pixel"] = "image",
        result_fit_range: Union[
            tuple[int, int, int, int],
            tuple[int, int, int, int, int, int],
            None,
        ] = None,
        result_input_arguments: Optional[Sequence[ParameterValues]] = None,
        target_fit_range: Union[
            tuple[int, int, int, int],
            tuple[int, int, int, int, int, int],
            None,
        ] = None,
        pygmo_seed: Optional[int] = None,
        pipeline_seed: Optional[int] = None,
        num_islands: int = 1,
        num_evolutions: int = 1,
        num_best_decisions: Optional[int] = None,
        topology: Literal["unconnected", "ring", "fully_connected"] = "unconnected",
        type_islands: Literal[
            "multiprocessing", "multithreading", "ipyparallel"
        ] = "multiprocessing",
        weights_from_file: Optional[Sequence[Union[str, Path]]] = None,
        weights: Optional[Sequence[float]] = None,
    ):
        if pygmo_seed is not None and pygmo_seed not in range(100001):
            raise ValueError("'Pygmo seed' must be between 0 and 100000.")

        if num_islands < 1:
            raise ValueError("'num_islands' must superior or equal to 1.")

        self._log = logging.getLogger(__name__)

        self.outputs = outputs
        self.readout: Readout = readout or Readout()

        self._calibration_mode = CalibrationMode(mode)

        self._result_type: ResultId = get_result_id(result_type)

        self._result_fit_range: Sequence[int] = result_fit_range or []

        self._result_input_arguments: Sequence[ParameterValues] = (
            result_input_arguments or []
        )

        self._target_data_path: Sequence[Path] = (
            to_path_list(target_data_path) if target_data_path else []
        )
        self._target_fit_range: Sequence[int] = target_fit_range or []

        self._fitness_function: FitnessFunction = fitness_function
        self._algorithm: Algorithm = algorithm

        self._parameters: Sequence[ParameterValues] = parameters or []

        if pygmo_seed is None:
            rng = np.random.default_rng()
            self._pygmo_seed: int = rng.integers(100000)
        else:
            self._pygmo_seed = pygmo_seed

        self._num_islands: int = num_islands
        self._num_evolutions: int = num_evolutions
        self._num_best_decisions: Optional[int] = num_best_decisions
        self._type_islands: Island = Island(type_islands)
        self._pipeline_seed = pipeline_seed
        self._topology: Literal["unconnected", "ring", "fully_connected"] = topology

        if weights and weights_from_file:
            raise ValueError("Cannot define both weights and weights from file.")

        self._weights_from_file: Optional[Sequence[Path]] = (
            to_path_list(weights_from_file) if weights_from_file else None
        )
        self._weights: Optional[Sequence[float]] = weights

    @property
    def calibration_mode(self) -> CalibrationMode:
        """TBW."""
        return self._calibration_mode

    @calibration_mode.setter
    def calibration_mode(self, value: CalibrationMode) -> None:
        """TBW."""
        self._calibration_mode = value

    @property
    def result_type(self) -> ResultId:
        """TBW."""
        return self._result_type

    @result_type.setter
    def result_type(self, value: ResultId) -> None:
        """TBW."""
        self._result_type = value

    @property
    def result_fit_range(self) -> Sequence[int]:
        """TBW."""
        return self._result_fit_range

    @result_fit_range.setter
    def result_fit_range(self, value: Sequence[int]) -> None:
        """TBW."""
        self._result_fit_range = value

    @property
    def result_input_arguments(self) -> Sequence[ParameterValues]:
        """TBW."""
        return self._result_input_arguments

    @result_input_arguments.setter
    def result_input_arguments(self, value: Sequence[ParameterValues]) -> None:
        """TBW."""
        self._result_input_arguments = value

    @property
    def target_data_path(self) -> Sequence[Path]:
        """TBW."""
        return self._target_data_path

    @target_data_path.setter
    def target_data_path(self, value: Sequence[Path]) -> None:
        """TBW."""
        self._target_data_path = value

    @property
    def target_fit_range(self) -> Sequence[int]:
        """TBW."""
        return self._target_fit_range

    @target_fit_range.setter
    def target_fit_range(self, value: Sequence[int]) -> None:
        """TBW."""
        self._target_fit_range = value

    @property
    def fitness_function(self) -> FitnessFunction:
        """TBW."""
        return self._fitness_function

    @fitness_function.setter
    def fitness_function(self, value: FitnessFunction) -> None:
        """TBW."""
        self._fitness_function = value

    @property
    def algorithm(self) -> Algorithm:
        """TBW."""
        return self._algorithm

    @algorithm.setter
    def algorithm(self, value: Algorithm) -> None:
        """TBW."""
        self._algorithm = value

    @property
    def parameters(self) -> Sequence[ParameterValues]:
        """TBW."""
        return self._parameters

    @parameters.setter
    def parameters(self, value: Sequence[ParameterValues]) -> None:
        """TBW."""
        self._parameters = value

    @property
    def pygmo_seed(self) -> int:
        """TBW."""
        return self._pygmo_seed

    @pygmo_seed.setter
    def pygmo_seed(self, value: int) -> None:
        """TBW."""
        if value not in range(100001):
            raise ValueError("Pygmo 'seed' must be between 0 and 100000.")

        self._pygmo_seed = value

    @property
    def pipeline_seed(self) -> Optional[int]:
        """TBW."""
        return self._pipeline_seed

    @pipeline_seed.setter
    def pipeline_seed(self, value: int) -> None:
        """TBW."""
        self._pipeline_seed = value

    @property
    def num_islands(self) -> int:
        """TBW."""
        return self._num_islands

    @num_islands.setter
    def num_islands(self, value: int) -> None:
        """TBW."""
        if value < 1:
            raise ValueError("'num_islands' must superior or equal to 1.")

        self._num_islands = value

    @property
    def num_evolutions(self) -> int:
        """TBW."""
        return self._num_evolutions

    @num_evolutions.setter
    def num_evolutions(self, value: int) -> None:
        """TBW."""
        self._num_evolutions = value

    @property
    def num_best_decisions(self) -> Optional[int]:
        """TBW."""
        return self._num_best_decisions

    @num_best_decisions.setter
    def num_best_decisions(self, value: Optional[int]) -> None:
        """TBW."""
        if isinstance(value, int) and value < 0:
            raise ValueError(
                "'num_best_decisions' must be 'None' or a positive integer"
            )

        self._num_best_decisions = value

    @property
    def topology(self) -> Literal["unconnected", "ring", "fully_connected"]:
        """TBW."""
        return self._topology

    @topology.setter
    def topology(self, value: Any) -> None:
        if value not in ("unconnected", "ring", "fully_connected"):
            raise ValueError(
                "Expecting value: 'unconnected', 'ring' or 'fully_connected'"
            )

        self._topology = value

    @property
    def weights_from_file(self) -> Optional[Sequence[Path]]:
        """TBW."""
        return self._weights_from_file

    @weights_from_file.setter
    def weights_from_file(self, value: Sequence[Path]) -> None:
        """TBW."""
        self._weights_from_file = value

    @property
    def weights(self) -> Optional[Sequence[float]]:
        """TBW."""
        return self._weights

    @weights.setter
    def weights(self, value: Sequence[float]) -> None:
        """TBW."""
        self._weights = value

    def get_problem(self, processor: Processor, output_dir: Path) -> ModelFitting:
        """Convert a 'processor' object into a Pygmo Problem.

        Examples
        --------
        Create a 'Pygmo Problem'
        >>> calibration = Calibration(...)
        >>> problem = calibration.get_problem(processor=..., output_dir=...)

        Create a decision vector
        >>> problem.get_bounds()
        >>> decision_vector = [...]

        Compute fitness
        >>> problem.fitness(decision_vector)
        """
        problem = ModelFitting(
            processor=processor,
            variables=self.parameters,
            readout=self.readout,
            calibration_mode=CalibrationMode(self.calibration_mode),
            simulation_output=self.result_type,
            generations=self.algorithm.generations,
            population_size=self.algorithm.population_size,
            fitness_func=self.fitness_function,
            file_path=output_dir,
        )

        problem.configure(
            target_output=self.target_data_path,
            target_fit_range=self.target_fit_range,
            out_fit_range=self.result_fit_range,
            input_arguments=self.result_input_arguments,
            weights=self.weights,
            weights_from_file=self.weights_from_file,
        )

        return problem

    # TODO: This function will be deprecated (see #563)
    def run_calibration(
        self,
        processor: Processor,
        output_dir: Path,
        with_progress_bar: bool = True,
    ) -> tuple["xr.Dataset", "pd.DataFrame", "pd.DataFrame"]:
        """Run calibration pipeline."""
        try:
            import pygmo as pg
        except ImportError as exc:
            raise ImportError(
                "Missing optional package 'pygmo'.\n"
                "Please install it with 'pip install pyxel-sim[calibration]' "
                "or 'pip install pyxel-sim[all]'"
            ) from exc

        pg.set_global_rng_seed(seed=self.pygmo_seed)
        self._log.info("Pygmo seed: %d", self.pygmo_seed)

        # Create a Pygmo problem
        fitting: ModelFitting = self.get_problem(
            processor=processor, output_dir=output_dir
        )

        # Create an archipelago
        user_defined_island = DaskIsland()
        user_defined_bfe = DaskBFE()

        if self.topology == "unconnected":
            topo = pg.unconnected()
        elif self.topology == "ring":
            topo = pg.ring()
        elif self.topology == "fully_connected":
            topo = pg.fully_connected()
        else:
            raise NotImplementedError(f"topology {self.topology!r}")

        # Create a new archipelago
        # This operation takes some time ...
        archipelago = MyArchipelago(
            num_islands=self.num_islands,
            udi=user_defined_island,
            algorithm=self.algorithm,
            problem=fitting,
            pop_size=self.algorithm.population_size,
            bfe=user_defined_bfe,
            topology=topo,
            pygmo_seed=self.pygmo_seed,
            with_bar=with_progress_bar,
        )

        # Run several evolutions in the archipelago
        ds, df_processors, df_all_logs = archipelago.run_evolve(
            readout=self.readout,
            num_evolutions=self._num_evolutions,
            num_best_decisions=self._num_best_decisions,
        )

        ds.attrs["topology"] = self.topology
        ds.attrs["result_type"] = str(fitting.sim_output)

        self._log.info("Calibration ended.")
        return ds, df_processors, df_all_logs

    def run_calibration_new(
        self,
        processor: Processor,
        output_dir: Path,
        with_progress_bar: bool = True,
    ) -> "DataTree":
        """Run calibration pipeline."""
        try:
            import pygmo as pg
        except ImportError as exc:
            raise ImportError(
                "Missing optional package 'pygmo'.\n"
                "Please install it with 'pip install pyxel-sim[calibration]' "
                "or 'pip install pyxel-sim[all]'"
            ) from exc

        pg.set_global_rng_seed(seed=self.pygmo_seed)
        self._log.info("Pygmo seed: %d", self.pygmo_seed)

        target_fit_range = to_fit_range(self.target_fit_range)
        result_fit_range = FitRange3D.from_sequence(self.result_fit_range)

        fitting: ModelFittingDataTree = ModelFittingDataTree(
            processor=processor,
            variables=self.parameters,
            readout=self.readout,
            simulation_output=self.result_type,
            generations=self.algorithm.generations,
            population_size=self.algorithm.population_size,
            fitness_func=self.fitness_function,
            file_path=output_dir,
            target_filenames=self.target_data_path,
            target_fit_range=target_fit_range,
            out_fit_range=result_fit_range,
            input_arguments=self.result_input_arguments,
            weights=self.weights,
            weights_from_file=self.weights_from_file,
        )

        # Create an archipelago
        user_defined_island = DaskIsland()
        user_defined_bfe = DaskBFE()

        if self.topology == "unconnected":
            topo = pg.unconnected()
        elif self.topology == "ring":
            topo = pg.ring()
        elif self.topology == "fully_connected":
            topo = pg.fully_connected()
        else:
            raise NotImplementedError(f"topology {self.topology!r}")

        # Create a new archipelago
        # This operation takes some time ...
        archipelago = ArchipelagoDataTree(
            num_islands=self.num_islands,
            udi=user_defined_island,
            algorithm=self.algorithm,
            problem=fitting,
            pop_size=self.algorithm.population_size,
            bfe=user_defined_bfe,
            topology=topo,
            pygmo_seed=self.pygmo_seed,
            with_bar=with_progress_bar,
        )

        # Run several evolutions in the archipelago
        data_tree: "DataTree" = archipelago.run_evolve(
            readout=self.readout,
            num_rows=processor.detector.geometry.row,
            num_cols=processor.detector.geometry.col,
            num_evolutions=self._num_evolutions,
            num_best_decisions=self._num_best_decisions,
        )

        data_tree.attrs["topology"] = self.topology
        data_tree.attrs["result_type"] = str(fitting.sim_output)

        self._log.info("Calibration ended.")
        return data_tree

    # TODO: This function will be deprecated (see #563)
    def _post_processing(
        self,
        ds: "xr.Dataset",
        df_processors: "pd.DataFrame",
        output: "CalibrationOutputs",
    ) -> Sequence[Delayed]:
        filenames: Sequence[Delayed] = output.save_processors(processors=df_processors)

        # TODO: Use output.fitting_plot ?
        # TODO: Use output.fitting_plot_close ?
        # TODO: Use output.calibration_plots ?

        return filenames

    @classmethod
    def from_json(cls, dct: Mapping) -> "Calibration":
        """Create a new object from a JSON dictionary."""
        from pyxel.outputs import CalibrationOutputs

        def _from_json(
            outputs: Mapping,
            target_data_path: Sequence[str],
            fitness_function: Mapping,
            algorithm: Mapping,
            parameters: Sequence[Mapping],
            readout: Optional[Mapping] = None,
            result_input_arguments: Optional[Sequence[Mapping]] = None,
            **kwargs,
        ) -> "Calibration":
            """Create a new object from an unpacked JSON dictionary."""
            return cls(
                outputs=CalibrationOutputs(**outputs),
                target_data_path=target_data_path,
                fitness_function=FitnessFunction(func=fitness_function["func"]),
                algorithm=Algorithm(**algorithm),
                parameters=[ParameterValues(**el) for el in parameters],
                readout=(Readout(**readout) if readout else None),
                result_input_arguments=(
                    [ParameterValues(**el) for el in result_input_arguments]
                    if result_input_arguments
                    else None
                ),
                **kwargs,
            )

        return _from_json(**dct)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import logging
from collections.abc import Iterator, Mapping, Sequence
from copy import deepcopy
from typing import TYPE_CHECKING

from pyxel.pipelines import ModelFunction

if TYPE_CHECKING:
    from pyxel.detectors import Detector


# TODO: These methods could also be as a `abc.Sequence` with magical methods:
#       __getitem__, __iter__, __len__, __contains__, ...
#       See #181
class ModelGroup:
    """TBW."""

    def __init__(self, models: Sequence[ModelFunction], name: str):
        self._log = logging.getLogger(__name__)
        self._name = name
        self.models: Sequence[ModelFunction] = models

    def __repr__(self):
        cls_name: str = self.__class__.__name__

        all_models: list[str] = [model.name for model in self.models if model.name]

        return f"{cls_name}<name={self._name!r}, models={all_models!r}>"

    def __deepcopy__(self, memo: dict) -> "ModelGroup":
        copied_models = deepcopy(self.models)
        return ModelGroup(models=copied_models, name=self._name)

    def __iter__(self) -> Iterator[ModelFunction]:
        for model in self.models:
            if model.enabled:
                yield model

    def __getstate__(self) -> Mapping:
        return {"models": tuple(self.models), "name": self._name}

    def __setstate__(self, state: Mapping) -> None:
        self.models = list(state["models"])
        self._name = state["name"]

    def __getattr__(self, item: str) -> ModelFunction:
        for model in self.models:
            if model.name == item:
                return model
        else:
            raise AttributeError(f"Cannot find model {item!r}.")

    def __dir__(self):
        return dir(type(self)) + [model.name for model in self.models]

    def run(
        self,
        detector: "Detector",
        with_intermediate_steps: bool = False,
    ):
        """Execute each enabled model in this group.

        Parameters
        ----------
        detector : Detector
        with_intermediate_steps : bool
        """
        model: ModelFunction
        for model in self:
            self._log.info("Model: %r", model.name)
            model(detector)

            # TODO: Refactor
            if with_intermediate_steps:
                import xarray as xr
                from datatree import DataTree

                # Get current absolute time
                absolute_time = xr.DataArray(
                    [detector.absolute_time],
                    dims=["time"],
                    attrs={"units": "s"},
                )

                # Get current Dataset
                absolute_time = xr.DataArray(absolute_time, attrs={"units": "s"})
                ds: xr.Dataset = detector.to_xarray().assign_coords(time=absolute_time)

                # Get datatree parent 'intermediate'
                intermediate_key: str = "intermediate"
                if intermediate_key not in detector.data:
                    datatree_intermediate: DataTree = DataTree(
                        name="intermediate", parent=detector.data
                    )
                    datatree_intermediate.attrs = {
                        "long_name": "Store all intermediate results modified along a pipeline"
                    }
                else:
                    datatree_intermediate = detector.data[intermediate_key]  # type: ignore

                # TODO: Refactor
                pipeline_key: str = f"idx_{detector.pipeline_count}"
                if pipeline_key not in datatree_intermediate:
                    datatree_one_time: DataTree = DataTree(
                        name=pipeline_key, parent=datatree_intermediate
                    )
                    datatree_one_time.attrs = {
                        "long_name": "Pipeline for one unique time",
                        "pipeline_count": detector.pipeline_count,
                        "time": f"{detector.absolute_time} s",
                    }
                else:
                    datatree_one_time = datatree_intermediate[pipeline_key]  # type: ignore

                # TODO: Refactor
                model_group_key: str = self._name
                if model_group_key not in datatree_one_time:
                    datatree_group: DataTree = DataTree(
                        name=model_group_key, parent=datatree_one_time
                    )
                    datatree_group.attrs = {"long_name": f"Model group: {self._name!r}"}
                else:
                    datatree_group = datatree_one_time[model_group_key]  # type: ignore

                # TODO: Refactor
                model_key: str = model.name
                if model_key not in datatree_group:
                    datatree_model: DataTree = DataTree(
                        name=model_key, parent=datatree_group
                    )
                    datatree_model.attrs = {"long_name": f"Group: {model.name!r}"}
                else:
                    datatree_model = datatree_group[model_key]  # type: ignore

                # TODO: Refactor
                last_key: str = "last"
                if last_key not in datatree_intermediate:
                    last_full_ds: xr.Dataset = xr.zeros_like(ds)
                else:
                    last_full_ds = datatree_intermediate[last_key]  # type: ignore

                for name, data_array in ds.data_vars.items():
                    previous_data_array = last_full_ds[name]

                    if not data_array.equals(previous_data_array):
                        datatree_model[name] = data_array

                # datatree_model
                datatree_intermediate[last_key] = DataTree(ds.copy(deep=True))



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel examples downloader."""

import shutil
from pathlib import Path
from zipfile import ZipFile

import requests
from tqdm import tqdm


def download_examples(foldername: str = "pyxel-examples", force: bool = False) -> None:
    """Download and save examples from Pyxel Data Gitlab repository in the working directory.

    Parameters
    ----------
    foldername: str
    force: bool
    """
    folder: Path = Path(foldername).resolve()

    if folder.is_dir():
        if force:
            shutil.rmtree(folder)
        else:
            raise OSError(
                f"Folder '{folder}' already exists. Either delete it, "
                f"use a different name, or use the '--force' argument."
            )

    url = "https://gitlab.com/esa/pyxel-data/-/archive/master/pyxel-data-master.zip"
    response = requests.get(url, stream=True)

    examples_filename = Path("examples_tmp.zip")

    total = int(response.headers.get("content-length", 0))
    with examples_filename.open("wb") as tmp, tqdm(
        desc="Downloading examples",
        total=total,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in response.iter_content(chunk_size=1024):
            size = tmp.write(data)
            bar.update(size)

    with ZipFile(examples_filename, "r") as zipobj:
        zipobj.extractall(folder)

    pyxel_data_folder: Path = folder.joinpath("pyxel-data-master")

    filename: Path
    for filename in pyxel_data_folder.glob("*"):
        relative_filename: Path = filename.relative_to(pyxel_data_folder)
        shutil.move(src=filename, dst=folder.joinpath(relative_filename))

    pyxel_data_folder.rmdir()
    examples_filename.unlink()

    print("Done.")
